"""Add reply_id to notifications.

Revision ID: 77d6820f494f
Revises: 87efceb6745b
Create Date: 2021-11-14 16:33:29.498193

"""

from alembic import op
import sqlalchemy as sa
import sqlalchemy_utils
import sqlalchemy_utils
import uuid

# revision identifiers, used by Alembic.
revision = "77d6820f494f"
down_revision = "87efceb6745b"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "notification",
        sa.Column(
            "reply_id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=True,
        ),
    )
    op.create_index(
        op.f("ix_notification_reply_id"),
        "notification",
        ["reply_id"],
        unique=False,
    )
    # ### end Alembic commands ###


def downgrade():
    op.drop_index(op.f("ix_notification_reply_id"), table_name="notification")
    op.drop_column("notification", "reply_id")
    # ### end Alembic commands ###



"""Add descriptions for entities tasks and statuses

Revision ID: a252a094e977
Revises: 1bb55759146f
Create Date: 2024-06-20 20:37:21.885953

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "a252a094e977"
down_revision = "1bb55759146f"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "task_type", sa.Column("description", sa.Text(), nullable=True)
    )
    op.add_column(
        "task_status", sa.Column("description", sa.Text(), nullable=True)
    )
    op.add_column(
        "entity_type", sa.Column("description", sa.Text(), nullable=True)
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("task_type", "description")
    op.drop_column("task_status", "description")
    op.drop_column("entity_type", "description")

    # ### end Alembic commands ###



"""Add Entity.is_shared

Revision ID: 59a7445a966c
Revises: ca28796a2a62
Create Date: 2024-08-16 17:31:59.331365

"""

from alembic import op
import sqlalchemy as sa
from sqlalchemy.sql import expression


# revision identifiers, used by Alembic.
revision = "59a7445a966c"
down_revision = "ca28796a2a62"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("entity", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "is_shared",
                sa.Boolean(),
                nullable=False,
                server_default=expression.false(),
            )
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("entity", schema=None) as batch_op:
        batch_op.drop_column("is_shared")

    # ### end Alembic commands ###



from sqlalchemy_utils import UUIDType
from zou.app import db
from zou.app.models.serializer import SerializerMixin
from zou.app.models.base import BaseMixin


class TimeSpent(db.Model, BaseMixin, SerializerMixin):
    """
    Describes the time spent by someone on a task.
    """

    duration = db.Column(db.Float, nullable=False)
    date = db.Column(db.Date, nullable=False)

    task_id = db.Column(
        UUIDType(binary=False), db.ForeignKey("task.id"), index=True
    )
    person_id = db.Column(
        UUIDType(binary=False), db.ForeignKey("person.id"), index=True
    )

    __table_args__ = (
        db.UniqueConstraint(
            "person_id", "task_id", "date", name="time_spent_uc"
        ),
        db.CheckConstraint("duration > 0", name="check_duration_positive"),
    )



from flask import Blueprint
from zou.app.utils.api import configure_api_from_blueprint

from zou.app.blueprints.news.resources import (
    NewsResource,
    ProjectNewsResource,
    ProjectSingleNewsResource,
)


routes = [
    ("/data/projects/news", NewsResource),
    ("/data/projects/<project_id>/news", ProjectNewsResource),
    ("/data/projects/<project_id>/news/<news_id>", ProjectSingleNewsResource),
]

blueprint = Blueprint("news", "news")
api = configure_api_from_blueprint(blueprint, routes)



from sqlalchemy_utils import UUIDType

from zou.app import db
from zou.app.models.serializer import SerializerMixin
from zou.app.models.base import BaseMixin
from zou.app.utils import fields


class Budget(db.Model, BaseMixin, SerializerMixin):
    """
    Budget quote for a project. It's a base object where budget entries
    are linked to.
    """

    project_id = db.Column(
        UUIDType(binary=False), db.ForeignKey("project.id"), index=True
    )
    revision = db.Column(db.Integer, nullable=False, default=1)
    name = db.Column(db.String(255), nullable=False)
    currency = db.Column(db.String(3))

    def __repr__(self):
        return "<Budget of %s - %d %s %s>" % (
            self.project_id,
            self.revision,
            self.name,
            self.id,
        )

    def present(self):
        return fields.serialize_dict(
            {
                "id": self.id,
                "project_id": self.project_id,
                "revision": self.revision,
                "name": self.name,
                "currency": self.currency,
            }
        )



"""TaskTypeAssetTypeLink/DepartmentMetadataDescriptorLink add index, pkey and constraint

Revision ID: 20a8ad264659
Revises: 539a3a00c417
Create Date: 2025-01-26 23:15:27.665887

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "20a8ad264659"
down_revision = "539a3a00c417"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table(
        "department_metadata_descriptor_link", schema=None
    ) as batch_op:
        batch_op.alter_column(
            "metadata_descriptor_id", existing_type=sa.UUID(), nullable=False
        )
        batch_op.alter_column(
            "department_id", existing_type=sa.UUID(), nullable=False
        )
        batch_op.create_unique_constraint(
            "department_metadata_descriptor_link_uc",
            ["metadata_descriptor_id", "department_id"],
        )
        batch_op.create_index(
            batch_op.f("ix_department_metadata_descriptor_link_department_id"),
            ["department_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f(
                "ix_department_metadata_descriptor_link_metadata_descriptor_id"
            ),
            ["metadata_descriptor_id"],
            unique=False,
        )

        batch_op.create_primary_key(
            "department_metadata_descriptor_link_pkey",
            ["metadata_descriptor_id", "department_id"],
        )

    with op.batch_alter_table(
        "task_type_asset_type_link", schema=None
    ) as batch_op:
        batch_op.alter_column(
            "asset_type_id", existing_type=sa.UUID(), nullable=False
        )
        batch_op.alter_column(
            "task_type_id", existing_type=sa.UUID(), nullable=False
        )
        batch_op.create_index(
            batch_op.f("ix_task_type_asset_type_link_asset_type_id"),
            ["asset_type_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_task_type_asset_type_link_task_type_id"),
            ["task_type_id"],
            unique=False,
        )
        batch_op.create_unique_constraint(
            "task_type_asset_type_link_uc", ["asset_type_id", "task_type_id"]
        )

        batch_op.create_primary_key(
            "task_type_asset_type_link_pkey", ["asset_type_id", "task_type_id"]
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table(
        "task_type_asset_type_link", schema=None
    ) as batch_op:
        batch_op.drop_constraint(
            "task_type_asset_type_link_pkey", type_="primary"
        )
        batch_op.drop_constraint(
            "task_type_asset_type_link_uc", type_="unique"
        )
        batch_op.drop_index(
            batch_op.f("ix_task_type_asset_type_link_task_type_id")
        )
        batch_op.drop_index(
            batch_op.f("ix_task_type_asset_type_link_asset_type_id")
        )

        batch_op.alter_column(
            "task_type_id", existing_type=sa.UUID(), nullable=True
        )
        batch_op.alter_column(
            "asset_type_id", existing_type=sa.UUID(), nullable=True
        )

    with op.batch_alter_table(
        "department_metadata_descriptor_link", schema=None
    ) as batch_op:
        batch_op.drop_index(
            batch_op.f(
                "ix_department_metadata_descriptor_link_metadata_descriptor_id"
            )
        )
        batch_op.drop_index(
            batch_op.f("ix_department_metadata_descriptor_link_department_id")
        )
        batch_op.drop_constraint(
            "department_metadata_descriptor_link_pkey", type_="primary"
        )
        batch_op.drop_constraint(
            "department_metadata_descriptor_link_uc", type_="unique"
        )
        batch_op.alter_column(
            "department_id", existing_type=sa.UUID(), nullable=True
        )
        batch_op.alter_column(
            "metadata_descriptor_id", existing_type=sa.UUID(), nullable=True
        )

    # ### end Alembic commands ###



"""Add validation status column to preview files

Revision ID: 398150912a3f
Revises: 3b0d1321079e
Create Date: 2021-11-09 00:07:04.543076

"""

from alembic import op
import sqlalchemy as sa
import sqlalchemy_utils


# revision identifiers, used by Alembic.
revision = "398150912a3f"
down_revision = "3b0d1321079e"
branch_labels = None
depends_on = None

VALIDATION_STATUSES = [
    ("validated", "Validated"),
    ("rejected", "Rejected"),
    ("neutral", "Neutral"),
]


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "preview_file",
        sa.Column(
            "validation_status",
            sqlalchemy_utils.types.choice.ChoiceType(VALIDATION_STATUSES),
            nullable=True,
        ),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("preview_file", "validation_status")
    # ### end Alembic commands ###



from sqlalchemy_utils import UUIDType

from zou.app import db
from zou.app.models.serializer import SerializerMixin
from zou.app.models.base import BaseMixin


class TaskType(db.Model, BaseMixin, SerializerMixin):
    """
    Categorize tasks in domain areas: modeling, animation, etc.
    """

    name = db.Column(db.String(40), nullable=False)
    short_name = db.Column(db.String(20))
    description = db.Column(db.Text())
    color = db.Column(db.String(7), default="#FFFFFF")
    priority = db.Column(db.Integer, default=1)
    for_entity = db.Column(db.String(30), default="Asset")
    allow_timelog = db.Column(db.Boolean, default=True)
    archived = db.Column(db.Boolean(), default=False)
    shotgun_id = db.Column(db.Integer, index=True)

    department_id = db.Column(
        UUIDType(binary=False), db.ForeignKey("department.id"), index=True
    )

    __table_args__ = (
        db.UniqueConstraint(
            "name", "for_entity", "department_id", name="task_type_uc"
        ),
    )



"""Add TaskStatus.priority

Revision ID: 693cc511d28d
Revises: 328fd44c6347
Create Date: 2024-01-11 15:41:29.314852

"""

from alembic import op
import sqlalchemy as sa
import sqlalchemy_utils
import uuid
from sqlalchemy.orm.session import Session
from sqlalchemy.ext.declarative import declarative_base
from zou.migrations.utils.base import BaseMixin


# revision identifiers, used by Alembic.
revision = "693cc511d28d"
down_revision = "328fd44c6347"
branch_labels = None
depends_on = None

base = declarative_base()


class ProjectTaskStatusLink(base, BaseMixin):
    __tablename__ = "project_task_status_link"
    project_id = sa.Column(
        sqlalchemy_utils.types.UUIDType(binary=False),
        primary_key=True,
    )
    task_status_id = sa.Column(
        sqlalchemy_utils.types.UUIDType(binary=False),
        primary_key=True,
    )
    priority = sa.Column(sa.Integer, default=None)

    __table_args__ = (
        sa.UniqueConstraint(
            "project_id", "task_status_id", name="project_taskstatus_uc"
        ),
    )


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table(
        "project_task_status_link", schema=None
    ) as batch_op:
        batch_op.add_column(sa.Column("priority", sa.Integer(), nullable=True))
        batch_op.add_column(
            sa.Column("created_at", sa.DateTime(), nullable=True)
        )
        batch_op.add_column(
            sa.Column("updated_at", sa.DateTime(), nullable=True)
        )

        batch_op.add_column(
            sa.Column(
                "id",
                sqlalchemy_utils.types.uuid.UUIDType(binary=False),
                default=uuid.uuid4,
                nullable=True,
            )
        )

    with op.batch_alter_table(
        "project_task_status_link", schema=None
    ) as batch_op:
        batch_op.execute(
            f"UPDATE project_task_status_link SET id = '{uuid.uuid4()}'"
        )

        session = Session(bind=op.get_bind())
        for link in session.query(ProjectTaskStatusLink).all():
            link.id = uuid.uuid4()
            session.add(link)
            session.commit()
        op.alter_column("project_task_status_link", "id", nullable=False)

        batch_op.create_unique_constraint(
            "project_taskstatus_uc", ["project_id", "task_status_id"]
        )

    with op.batch_alter_table("task_status", schema=None) as batch_op:
        batch_op.add_column(sa.Column("priority", sa.Integer(), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("task_status", schema=None) as batch_op:
        batch_op.drop_column("priority")

    with op.batch_alter_table(
        "project_task_status_link", schema=None
    ) as batch_op:
        batch_op.drop_constraint("project_taskstatus_uc", type_="unique")
        batch_op.drop_column("priority")
        batch_op.drop_column("updated_at")
        batch_op.drop_column("created_at")
        batch_op.drop_column("id")

    # ### end Alembic commands ###



from zou.app.models.plugin import Plugin

from zou.app.blueprints.crud.base import BaseModelsResource, BaseModelResource


class PluginsResource(BaseModelsResource):
    def __init__(self):
        BaseModelsResource.__init__(self, Plugin)


class PluginResource(BaseModelResource):
    def __init__(self):
        BaseModelResource.__init__(self, Plugin)



"""empty message

Revision ID: af1790868e2c
Revises: a519c710877c
Create Date: 2019-02-26 20:36:21.400109

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "af1790868e2c"
down_revision = "a519c710877c"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "entity", sa.Column("nb_frames", sa.Integer(), nullable=True)
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("entity", "nb_frames")
    # ### end Alembic commands ###



"""add department mentions to comments

Revision ID: 0596674df51d
Revises: 9e5b3a9b0cee
Create Date: 2023-06-29 20:14:06.510015

"""

from alembic import op
import sqlalchemy as sa
import sqlalchemy_utils
import sqlalchemy_utils
import uuid

# revision identifiers, used by Alembic.
revision = "0596674df51d"
down_revision = "9e5b3a9b0cee"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "comment_department_mentions",
        sa.Column(
            "comment",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=False,
        ),
        sa.Column(
            "department",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["comment"],
            ["comment.id"],
        ),
        sa.ForeignKeyConstraint(
            ["department"],
            ["department.id"],
        ),
        sa.PrimaryKeyConstraint("comment", "department"),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("comment_department_mentions")
    # ### end Alembic commands ###



from flask import current_app

from zou.app.models.project import Project
from zou.app.models.entity import Entity

from zou.app.services import shots_service, persons_service

from zou.app.blueprints.source.shotgun.base import (
    BaseImportShotgunResource,
    ImportRemoveShotgunBaseResource,
)


class ImportShotgunShotsResource(BaseImportShotgunResource):
    def __init__(self):
        BaseImportShotgunResource.__init__(self)

    def prepare_import(self):
        self.shot_type = shots_service.get_shot_type()
        self.project_map = Project.get_id_map(field="name")
        self.current_user_id = persons_service.get_current_user()["id"]

    def extract_status_names(self, sg_projects):
        return {x["sg_status"] for x in sg_projects}

    def extract_data(self, sg_shot):
        (frame_in, frame_out) = self.extract_frame_range(sg_shot)
        custom_fields = self.extract_custom_data(sg_shot)
        project_id = self.get_project(sg_shot, self.project_map)
        sequence_id = self.get_sequence(sg_shot)
        scene_id = self.get_scene(sg_shot)
        assets = self.extract_assets(sg_shot)

        shot_type = shots_service.get_shot_type()

        data = {
            "name": sg_shot["code"],
            "shotgun_id": sg_shot["id"],
            "project_id": project_id,
            "entity_type_id": shot_type["id"],
            "parent_id": sequence_id,
            "source_id": scene_id,
            "entities_out": assets,
        }
        data_field_content = {"frame_in": frame_in, "frame_out": frame_out}
        custom_fields.update(data_field_content)
        data["data"] = custom_fields
        return data

    def get_project(self, sg_shot, project_map):
        project_id = None
        if sg_shot["project"] is not None:
            project_id = project_map.get(sg_shot["project"]["name"], None)
        return project_id

    def get_scene(self, sg_shot):
        scene_id = None
        if "sg_scene" in sg_shot and sg_shot["sg_scene"] is not None:
            scene_id = self.get_scene_id(sg_shot["sg_scene"]["id"])
        return scene_id

    def get_sequence(self, sg_shot):
        sequence_id = None
        if "sg_sequence" in sg_shot and sg_shot["sg_sequence"] is not None:
            sequence_id = self.get_sequence_id(sg_shot["sg_sequence"]["id"])
        return sequence_id

    def extract_frame_range(self, sg_shot):
        frame_in = sg_shot["sg_cut_in"]
        frame_out = frame_in
        if frame_in is not None and sg_shot["sg_cut_duration"] is not None:
            frame_out = sg_shot["sg_cut_in"] + sg_shot["sg_cut_duration"]
        return (frame_in, frame_out)

    def extract_assets(self, sg_shot):
        assets = []
        if "assets" in sg_shot and len(sg_shot["assets"]) > 0:
            for sg_asset in sg_shot["assets"]:
                entity_id = self.get_asset_id(sg_asset["id"])
                if entity_id is not None:
                    asset = Entity.get(entity_id)
                    assets.append(asset)
        return assets

    def is_custom_field(self, name):
        non_custom_fields = ["sg_cut_in", "sg_cut_out", "sg_sequence"]
        return name[:3] == "sg_" and name not in non_custom_fields

    def import_entry(self, data):
        shot = Entity.get_by(
            shotgun_id=data["shotgun_id"],
            entity_type_id=shots_service.get_shot_type()["id"],
        )

        if shot is None:
            shot = Entity.create(**data, created_by=self.current_user_id)
            current_app.logger.info("Shot created: %s" % shot)

        else:
            if shot.data is None:
                shot.data = {}
            shot.update(data)
            shot.data.update(data["data"])
            shot.save()
            shots_service.clear_shot_cache(str(shot.id))
            current_app.logger.info("Shot updated: %s" % shot)

        return shot


class ImportRemoveShotgunShotResource(ImportRemoveShotgunBaseResource):
    def __init__(self):
        ImportRemoveShotgunBaseResource.__init__(
            self, Entity, entity_type_id=shots_service.get_shot_type()["id"]
        )



"""Disallow null ChoiceType

Revision ID: 17ef8f7be758
Revises: 45dafbb3f4e1
Create Date: 2024-03-01 00:18:48.971796

"""

from alembic import op
import sqlalchemy as sa
from sqlalchemy.orm.session import Session
from zou.migrations.utils.base import BaseMixin
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy_utils import ChoiceType, EmailType, LocaleType, TimezoneType
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy_utils import UUIDType
from babel import Locale
from pytz import timezone as pytz_timezone


# revision identifiers, used by Alembic.
revision = "17ef8f7be758"
down_revision = "45dafbb3f4e1"
branch_labels = None
depends_on = None

base = declarative_base()

PROJECT_STYLES = [
    ("2d", "2D Animation"),
    ("3d", "3D Animation"),
    ("2d3d", "2D/3D Animation"),
    ("ar", "Augmented Reality"),
    ("vfx", "VFX"),
    ("stop-motion", "Stop Motion"),
    ("motion-design", "Motion Design"),
    ("archviz", "Archviz"),
    ("commercial", "Commercial"),
    ("catalog", "Catalog"),
    ("immersive", "Immersive Experience"),
    ("nft", "NFT Collection"),
    ("video-game", "Video Game"),
    ("vr", "Virtual Reality"),
]


class Project(base, BaseMixin):
    """
    Describes a production the studio works on.
    """

    __tablename__ = "project"

    name = sa.Column(sa.String(80), nullable=False, unique=True, index=True)
    code = sa.Column(sa.String(80))
    description = sa.Column(sa.Text())
    shotgun_id = sa.Column(sa.Integer)
    file_tree = sa.Column(JSONB)
    data = sa.Column(JSONB)
    has_avatar = sa.Column(sa.Boolean(), default=False)
    fps = sa.Column(sa.String(10), default=25)
    ratio = sa.Column(sa.String(10), default="16:9")
    resolution = sa.Column(sa.String(12), default="1920x1080")
    production_type = sa.Column(sa.String(20), default="short")
    production_style = sa.Column(
        ChoiceType(PROJECT_STYLES), default="2d3d", nullable=False
    )
    start_date = sa.Column(sa.Date())
    end_date = sa.Column(sa.Date())
    man_days = sa.Column(sa.Integer)
    nb_episodes = sa.Column(sa.Integer, default=0)
    episode_span = sa.Column(sa.Integer, default=0)
    max_retakes = sa.Column(sa.Integer, default=0)
    is_clients_isolated = sa.Column(sa.Boolean(), default=False)
    is_preview_download_allowed = sa.Column(sa.Boolean(), default=False)
    is_set_preview_automated = sa.Column(sa.Boolean(), default=False)
    homepage = sa.Column(sa.String(80), default="assets")


ENTITY_STATUSES = [
    ("standby", "Stand By"),
    ("running", "Running"),
    ("complete", "Complete"),
    ("canceled", "Canceled"),
]


STATUSES = [
    ("processing", "Processing"),
    ("ready", "Ready"),
    ("broken", "Broken"),
]

VALIDATION_STATUSES = [
    ("validated", "Validated"),
    ("rejected", "Rejected"),
    ("neutral", "Neutral"),
]


class PreviewFile(base, BaseMixin):
    """
    Describes a file which is aimed at being reviewed. It is not a publication
    neither a working file.
    """

    __tablename__ = "preview_file"
    name = sa.Column(sa.String(250))
    original_name = sa.Column(sa.String(250))
    revision = sa.Column(sa.Integer(), default=1)
    position = sa.Column(sa.Integer(), default=1)
    extension = sa.Column(sa.String(6))
    description = sa.Column(sa.Text())
    path = sa.Column(sa.String(400))
    source = sa.Column(sa.String(40))
    file_size = sa.Column(sa.BigInteger(), default=0)
    status = sa.Column(ChoiceType(STATUSES), default="processing")
    validation_status = sa.Column(
        ChoiceType(VALIDATION_STATUSES), default="neutral", nullable=False
    )
    annotations = sa.Column(JSONB)
    width = sa.Column(sa.Integer(), default=0)
    height = sa.Column(sa.Integer(), default=0)
    duration = sa.Column(sa.Float, default=0)

    task_id = sa.Column(
        UUIDType(binary=False), sa.ForeignKey("task.id"), index=True
    )
    person_id = sa.Column(UUIDType(binary=False), sa.ForeignKey("person.id"))
    source_file_id = sa.Column(
        UUIDType(binary=False), sa.ForeignKey("output_file.id")
    )

    __table_args__ = (
        sa.UniqueConstraint("name", "task_id", "revision", name="preview_uc"),
    )

    shotgun_id = sa.Column(sa.Integer, unique=True)

    is_movie = sa.Column(sa.Boolean, default=False)  # deprecated
    url = sa.Column(sa.String(600))  # deprecated
    uploaded_movie_url = sa.Column(sa.String(600))  # deprecated
    uploaded_movie_name = sa.Column(sa.String(150))  # deprecated


class Entity(base, BaseMixin):
    """
    Base model to represent assets, shots, sequences, episodes and scenes.
    They have different meaning but they share the same behaviour toward
    tasks and files.
    """

    __tablename__ = "entity"

    name = sa.Column(sa.String(160), nullable=False)
    code = sa.Column(sa.String(160))  # To store sanitized version of name
    description = sa.Column(sa.String(1200))
    shotgun_id = sa.Column(sa.Integer)
    canceled = sa.Column(sa.Boolean, default=False)

    nb_frames = sa.Column(sa.Integer)  # Specific to shots
    nb_entities_out = sa.Column(sa.Integer, default=0)
    is_casting_standby = sa.Column(sa.Boolean, default=False)

    status = sa.Column(
        ChoiceType(ENTITY_STATUSES), default="running", nullable=False
    )

    project_id = sa.Column(
        UUIDType(binary=False),
    )
    entity_type_id = sa.Column(
        UUIDType(binary=False),
    )

    parent_id = sa.Column(UUIDType(binary=False))  # sequence or episode

    source_id = sa.Column(
        UUIDType(binary=False),
    )  # if the entity is generated from another one (like shots from scene).

    preview_file_id = sa.Column(
        UUIDType(binary=False),
    )
    data = sa.Column(JSONB)

    ready_for = sa.Column(
        UUIDType(binary=False),
    )

    created_by = sa.Column(
        UUIDType(binary=False),
    )


CHANGE_TYPES = [("status", "Status"), ("ready_for", "Ready for")]


class StatusAutomation(
    base,
    BaseMixin,
):
    """
    Status automations are entries that allow to describe changes that
    should automatically apply after a task status is changed.

    For instance, a Modeling task set to done will imply to set the Rigging
    task status to ready and the *ready_for* field to be set at Layout.
    """

    __tablename__ = "status_automation"
    entity_type = sa.Column(sa.String(40), default="asset")

    in_task_type_id = sa.Column(
        UUIDType(binary=False), sa.ForeignKey("task_type.id"), index=True
    )
    in_task_status_id = sa.Column(
        UUIDType(binary=False), sa.ForeignKey("task_status.id"), index=True
    )

    out_field_type = sa.Column(
        ChoiceType(CHANGE_TYPES), default="status", nullable=False
    )
    out_task_type_id = sa.Column(
        UUIDType(binary=False), sa.ForeignKey("task_type.id"), index=True
    )
    out_task_status_id = sa.Column(
        UUIDType(binary=False),
        sa.ForeignKey("task_status.id"),
        index=True,
        nullable=True,
    )
    archived = sa.Column(sa.Boolean(), default=False)


STATUSES = [
    ("running", "Running"),
    ("failed", "Failed"),
    ("succeeded", "Succeeded"),
]

TYPES = [("archive", "Archive"), ("movie", "Movie")]

TWO_FACTOR_AUTHENTICATION_TYPES = [
    ("totp", "TOTP"),
    ("email_otp", "Email OTP"),
    ("fido", "FIDO"),
]

CONTRACT_TYPES = [
    ("open-ended", "Open-ended"),
    ("fixed-term", "Fixed-term"),
    ("short-term", "Short-term"),
    ("freelance", "Freelance"),
    ("apprentice", "Apprentice"),
    ("internship", "Internship"),
]

ROLE_TYPES = [
    ("user", "Artist"),
    ("admin", "Studio Manager"),
    ("supervisor", "Supervisor"),
    ("manager", "Production Manager"),
    ("client", "Client"),
    ("vendor", "Vendor"),
]


class Person(base, BaseMixin):
    """
    Describe a member of the studio (and an API user).
    """

    __tablename__ = "person"
    first_name = sa.Column(sa.String(80), nullable=False)
    last_name = sa.Column(sa.String(80), nullable=False)
    email = sa.Column(EmailType, unique=True)
    phone = sa.Column(sa.String(30))
    contract_type = sa.Column(ChoiceType(CONTRACT_TYPES), default="open-ended")

    active = sa.Column(sa.Boolean(), default=True)
    archived = sa.Column(sa.Boolean(), default=False)
    last_presence = sa.Column(sa.Date())

    password = sa.Column(sa.LargeBinary(60))
    desktop_login = sa.Column(sa.String(80))
    login_failed_attemps = sa.Column(sa.Integer, default=0)
    last_login_failed = sa.Column(sa.DateTime())
    totp_enabled = sa.Column(sa.Boolean(), default=False)
    totp_secret = sa.Column(sa.String(32), default=None)
    email_otp_enabled = sa.Column(sa.Boolean(), default=False)
    email_otp_secret = sa.Column(sa.String(32), default=None)
    fido_enabled = sa.Column(sa.Boolean(), default=False)
    fido_credentials = sa.Column(sa.ARRAY(JSONB))
    otp_recovery_codes = sa.Column(sa.ARRAY(sa.LargeBinary(60)))
    preferred_two_factor_authentication = sa.Column(
        ChoiceType(TWO_FACTOR_AUTHENTICATION_TYPES)
    )

    shotgun_id = sa.Column(sa.Integer, unique=True)
    timezone = sa.Column(
        TimezoneType(backend="pytz"),
        default=pytz_timezone("Europe/Paris"),
    )
    locale = sa.Column(LocaleType, default=Locale("en", "US"))
    data = sa.Column(JSONB)
    role = sa.Column(ChoiceType(ROLE_TYPES), default="user")
    has_avatar = sa.Column(sa.Boolean(), default=False)

    notifications_enabled = sa.Column(sa.Boolean(), default=False)
    notifications_slack_enabled = sa.Column(sa.Boolean(), default=False)
    notifications_slack_userid = sa.Column(sa.String(60), default="")
    notifications_mattermost_enabled = sa.Column(sa.Boolean(), default=False)
    notifications_mattermost_userid = sa.Column(sa.String(60), default="")
    notifications_discord_enabled = sa.Column(sa.Boolean(), default=False)
    notifications_discord_userid = sa.Column(sa.String(60), default="")

    is_bot = sa.Column(sa.Boolean(), default=False, nullable=False)
    jti = sa.Column(sa.String(60), nullable=True, unique=True)
    expiration_date = sa.Column(sa.Date(), nullable=True)


class BuildJob(
    base,
    BaseMixin,
):
    """
    A build job stores information about the state of the building
    of a given playlist.
    """

    __tablename__ = "build_job"

    status = sa.Column(ChoiceType(STATUSES), default="running", nullable=False)
    job_type = sa.Column(ChoiceType(TYPES), default="movie", nullable=False)
    ended_at = sa.Column(sa.DateTime)

    playlist_id = sa.Column(
        UUIDType(binary=False),
        sa.ForeignKey("playlist.id"),
        nullable=False,
        index=True,
    )


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("build_job", schema=None) as batch_op:
        session = Session(bind=op.get_bind())
        session.query(BuildJob).where(BuildJob.status == None).update(
            {
                BuildJob.status: "running",
            }
        )
        session.query(BuildJob).where(BuildJob.job_type == None).update(
            {
                BuildJob.job_type: "movie",
            }
        )
        session.commit()
        batch_op.alter_column(
            "status", existing_type=sa.VARCHAR(length=255), nullable=False
        )
        batch_op.alter_column(
            "job_type", existing_type=sa.VARCHAR(length=255), nullable=False
        )

    with op.batch_alter_table("entity", schema=None) as batch_op:
        session = Session(bind=op.get_bind())
        session.query(Entity).where(Entity.status == None).update(
            {
                Entity.status: "running",
            }
        )
        session.commit()
        batch_op.alter_column(
            "status", existing_type=sa.VARCHAR(length=255), nullable=False
        )

    with op.batch_alter_table("person", schema=None) as batch_op:
        session = Session(bind=op.get_bind())
        session.query(Person).where(Person.role == None).update(
            {
                Person.role: "user",
            }
        )
        session.commit()
        batch_op.alter_column(
            "role", existing_type=sa.VARCHAR(length=255), nullable=False
        )

    with op.batch_alter_table("preview_file", schema=None) as batch_op:
        session = Session(bind=op.get_bind())
        session.query(PreviewFile).where(
            PreviewFile.validation_status == None
        ).update(
            {
                PreviewFile.validation_status: "neutral",
            }
        )
        session.query(PreviewFile).where(PreviewFile.status == None).update(
            {
                PreviewFile.status: "processing",
            }
        )
        session.commit()
        batch_op.alter_column(
            "status", existing_type=sa.VARCHAR(length=255), nullable=False
        )
        batch_op.alter_column(
            "validation_status",
            existing_type=sa.VARCHAR(length=255),
            nullable=False,
        )

    with op.batch_alter_table("project", schema=None) as batch_op:
        session = Session(bind=op.get_bind())
        session.query(Project).where(Project.production_style == None).update(
            {
                Project.production_style: "2d3d",
            }
        )
        session.commit()
        batch_op.alter_column(
            "production_style",
            existing_type=sa.VARCHAR(length=255),
            nullable=False,
        )

    with op.batch_alter_table("status_automation", schema=None) as batch_op:
        session = Session(bind=op.get_bind())
        session.query(StatusAutomation).where(
            StatusAutomation.out_field_type == None
        ).update(
            {
                StatusAutomation.out_field_type: "status",
            }
        )
        session.commit()
        batch_op.alter_column(
            "out_field_type",
            existing_type=sa.VARCHAR(length=255),
            nullable=False,
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("status_automation", schema=None) as batch_op:
        batch_op.alter_column(
            "out_field_type",
            existing_type=sa.VARCHAR(length=255),
            nullable=True,
        )

    with op.batch_alter_table("project", schema=None) as batch_op:
        batch_op.alter_column(
            "production_style",
            existing_type=sa.VARCHAR(length=255),
            nullable=True,
        )

    with op.batch_alter_table("preview_file", schema=None) as batch_op:
        batch_op.alter_column(
            "validation_status",
            existing_type=sa.VARCHAR(length=255),
            nullable=True,
        )

    with op.batch_alter_table("person", schema=None) as batch_op:
        batch_op.alter_column(
            "role", existing_type=sa.VARCHAR(length=255), nullable=True
        )

    with op.batch_alter_table("entity", schema=None) as batch_op:
        batch_op.alter_column(
            "status", existing_type=sa.VARCHAR(length=255), nullable=True
        )

    with op.batch_alter_table("build_job", schema=None) as batch_op:
        batch_op.alter_column(
            "job_type", existing_type=sa.VARCHAR(length=255), nullable=True
        )
        batch_op.alter_column(
            "status", existing_type=sa.VARCHAR(length=255), nullable=True
        )

    # ### end Alembic commands ###



from zou.app.models.studio import Studio

from zou.app.blueprints.crud.base import BaseModelsResource, BaseModelResource

from zou.app.services import tasks_service


class StudiosResource(BaseModelsResource):
    def __init__(self):
        BaseModelsResource.__init__(self, Studio)

    def check_read_permissions(self, options=None):
        return True

    def post_creation(self, instance):
        tasks_service.clear_studio_cache(str(instance.id))
        return instance.serialize()


class StudioResource(BaseModelResource):
    def __init__(self):
        BaseModelResource.__init__(self, Studio)

    def check_read_permissions(self, instance):
        return True

    def post_update(self, instance_dict, data):
        tasks_service.clear_studio_cache(instance_dict["id"])
        return instance_dict

    def post_delete(self, instance_dict):
        tasks_service.clear_studio_cache(instance_dict["id"])
        return instance_dict



"""For description of entity, task, working_file don't set any length

Revision ID: f344b867a911
Revises: 32f134ff1201
Create Date: 2024-05-22 15:46:07.855818

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "f344b867a911"
down_revision = "32f134ff1201"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("entity", schema=None) as batch_op:
        batch_op.alter_column(
            "description",
            existing_type=sa.VARCHAR(length=1200),
            type_=sa.Text(),
            existing_nullable=True,
        )

    with op.batch_alter_table("task", schema=None) as batch_op:
        batch_op.alter_column(
            "description",
            existing_type=sa.VARCHAR(length=200),
            type_=sa.Text(),
            existing_nullable=True,
        )

    with op.batch_alter_table("working_file", schema=None) as batch_op:
        batch_op.alter_column(
            "description",
            existing_type=sa.VARCHAR(length=200),
            type_=sa.Text(),
            existing_nullable=True,
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("working_file", schema=None) as batch_op:
        batch_op.alter_column(
            "description",
            existing_type=sa.Text(),
            type_=sa.VARCHAR(length=200),
            existing_nullable=True,
        )

    with op.batch_alter_table("task", schema=None) as batch_op:
        batch_op.alter_column(
            "description",
            existing_type=sa.Text(),
            type_=sa.VARCHAR(length=200),
            existing_nullable=True,
        )

    with op.batch_alter_table("entity", schema=None) as batch_op:
        batch_op.alter_column(
            "description",
            existing_type=sa.Text(),
            type_=sa.VARCHAR(length=1200),
            existing_nullable=True,
        )

    # ### end Alembic commands ###



"""add day off table

Revision ID: 8e4f39e321f4
Revises: fb87feaaa094
Create Date: 2021-02-13 01:03:30.408954

"""

from alembic import op
import sqlalchemy as sa
import sqlalchemy_utils
import sqlalchemy_utils
import uuid

# revision identifiers, used by Alembic.
revision = "8e4f39e321f4"
down_revision = "fb87feaaa094"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "day_off",
        sa.Column(
            "id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=False,
        ),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("updated_at", sa.DateTime(), nullable=True),
        sa.Column("date", sa.Date(), nullable=False),
        sa.Column(
            "person_id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=True,
        ),
        sa.ForeignKeyConstraint(
            ["person_id"],
            ["person.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("person_id", "date", name="day_off_uc"),
    )
    op.create_index(
        op.f("ix_day_off_person_id"), "day_off", ["person_id"], unique=False
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_day_off_person_id"), table_name="day_off")
    op.drop_table("day_off")
    # ### end Alembic commands ###



import nomad
import base64
import orjson as json
import textwrap
import time


def run_job(app, config, nomad_job_name, params):
    nomad_host = config.JOB_QUEUE_NOMAD_HOST

    params.update(
        {
            k: v
            for k, v in config.__dict__.items()
            if k.startswith("FS_") and v is not None
        }
    )

    data = json.dumps(params)
    payload = base64.b64encode(data).decode("utf-8")
    ncli = nomad.Nomad(host=nomad_host, timeout=5)

    response = ncli.job.dispatch_job(nomad_job_name, payload=payload)

    nomad_jobid = response["DispatchedJobID"]

    while True:
        summary = ncli.job.get_summary(nomad_jobid)
        task_group = list(summary["Summary"])[0]
        status = summary["Summary"][task_group]
        if status["Failed"] != 0 or status["Lost"] != 0:
            app.logger.error("Nomad job %r failed: %r", nomad_jobid, status)
            out, err = get_nomad_job_logs(ncli, nomad_jobid, nomad_job_name)
            out = textwrap.indent(out, "\t")
            err = textwrap.indent(err, "\t")
            raise Exception(
                "Job %s is 'Failed' or 'Lost':\nStatus: "
                "%s\nerr:\n%s\nout:\n%s" % (nomad_jobid, status, err, out)
            )
            return False
        if status["Complete"] == 1:
            app.logger.info("Nomad job %r: complete", nomad_jobid)
            break
        # there isn't a timeout here but python rq jobs have a timeout. Nomad
        # jobs have a timeout too.
        time.sleep(1)
    return True


def get_nomad_job_logs(ncli, nomad_jobid, nomad_job_name):
    allocations = ncli.job.get_allocations(nomad_jobid)
    last = max(
        [(alloc["CreateIndex"], idx) for idx, alloc in enumerate(allocations)]
    )[1]
    alloc_id = allocations[last]["ID"]
    # logs aren't available when the task isn't started
    task = allocations[last]["TaskStates"][nomad_job_name]
    if not task["StartedAt"]:
        out = "\n".join([x["DisplayMessage"] for x in task["Events"]])
        err = ""
    else:
        err = ncli.client.stream_logs.stream(
            alloc_id, nomad_job_name, "stderr"
        )
        out = ncli.client.stream_logs.stream(
            alloc_id, nomad_job_name, "stdout"
        )
        if err:
            err = json.loads(err).get("Data", "")
            err = base64.b64decode(err).decode("utf-8")
        if out:
            out = json.loads(out).get("Data", "")
            out = base64.b64decode(out).decode("utf-8")
    return out.rstrip(), err.rstrip()



"""add person id to shot history

Revision ID: e839d6603c09
Revises: 346250b5304c
Create Date: 2020-12-14 12:00:19.045783

"""

from alembic import op
import sqlalchemy as sa
import sqlalchemy_utils
import sqlalchemy_utils
import uuid

# revision identifiers, used by Alembic.
revision = "e839d6603c09"
down_revision = "346250b5304c"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "entity_version",
        sa.Column(
            "person_id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=True,
        ),
    )
    op.create_index(
        op.f("ix_entity_version_person_id"),
        "entity_version",
        ["person_id"],
        unique=False,
    )
    op.create_foreign_key(
        None, "entity_version", "person", ["person_id"], ["id"]
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(None, "entity_version", type_="foreignkey")
    op.drop_index(
        op.f("ix_entity_version_person_id"), table_name="entity_version"
    )
    op.drop_column("entity_version", "person_id")
    # ### end Alembic commands ###



from zou.app.models.task_status import TaskStatus
from zou.app.services import tasks_service
from zou.app.blueprints.crud.base import BaseModelResource, BaseModelsResource


class TaskStatusesResource(BaseModelsResource):
    def __init__(self):
        BaseModelsResource.__init__(self, TaskStatus)

    def check_read_permissions(self, options=None):
        return True

    def post_creation(self, instance):
        tasks_service.clear_task_status_cache(str(instance.id))
        return instance.serialize()

    def check_creation_integrity(self, data):
        if data.get("is_default", False):
            status = TaskStatus.get_by(is_default=True)
            if status:
                status.update({"is_default": False})
        return data


class TaskStatusResource(BaseModelResource):
    def __init__(self):
        BaseModelResource.__init__(self, TaskStatus)

    def check_read_permissions(self, instance):
        return True

    def pre_update(self, instance_dict, data):
        if data.get("is_default", False):
            status = TaskStatus.get_by(is_default=True)
            if status:
                status.update({"is_default": False})
        return instance_dict

    def post_update(self, instance_dict, data):
        tasks_service.clear_task_status_cache(instance_dict["id"])
        return instance_dict

    def post_delete(self, instance_dict):
        tasks_service.clear_task_status_cache(instance_dict["id"])
        return instance_dict



"""Add columns to searchFilter table to create search filters groups

Revision ID: 7748d3d22925
Revises: ae0127f2fc56
Create Date: 2023-09-14 17:50:53.903284

"""

from alembic import op
import sqlalchemy as sa
import sqlalchemy_utils
import sqlalchemy_utils
import uuid

# revision identifiers, used by Alembic.
revision = "7748d3d22925"
down_revision = "ae0127f2fc56"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "search_filter_group",
        sa.Column("list_type", sa.String(length=80), nullable=False),
        sa.Column("entity_type", sa.String(length=80), nullable=True),
        sa.Column("name", sa.String(length=200), nullable=False),
        sa.Column("color", sa.String(length=8), nullable=False),
        sa.Column(
            "person_id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=True,
        ),
        sa.Column(
            "project_id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=True,
        ),
        sa.Column(
            "id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=False,
        ),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("updated_at", sa.DateTime(), nullable=True),
        sa.ForeignKeyConstraint(
            ["person_id"],
            ["person.id"],
        ),
        sa.ForeignKeyConstraint(
            ["project_id"],
            ["project.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    with op.batch_alter_table("search_filter_group", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_search_filter_group_list_type"),
            ["list_type"],
            unique=False,
        )

    with op.batch_alter_table("search_filter", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "search_filter_group_id",
                sqlalchemy_utils.types.uuid.UUIDType(binary=False),
                default=uuid.uuid4,
                nullable=True,
            )
        )
        batch_op.create_foreign_key(
            None, "search_filter_group", ["search_filter_group_id"], ["id"]
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("search_filter", schema=None) as batch_op:
        batch_op.drop_constraint(None, type_="foreignkey")
        batch_op.drop_column("search_filter_group_id")

    with op.batch_alter_table("search_filter_group", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_search_filter_group_list_type"))

    op.drop_table("search_filter_group")
    # ### end Alembic commands ###



"""Add missing index

Revision ID: fba149993140
Revises: 20a8ad264659
Create Date: 2025-02-19 02:19:27.407568

"""

from alembic import op


# revision identifiers, used by Alembic.
revision = "fba149993140"
down_revision = "20a8ad264659"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("comment_preview_link", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_comment_preview_link_comment"),
            ["comment"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_comment_preview_link_preview_file"),
            ["preview_file"],
            unique=False,
        )

    with op.batch_alter_table("project_person_link", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_project_person_link_person_id"),
            ["person_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_project_person_link_project_id"),
            ["project_id"],
            unique=False,
        )

    with op.batch_alter_table(
        "project_status_automation_link", schema=None
    ) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_project_status_automation_link_project_id"),
            ["project_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f(
                "ix_project_status_automation_link_status_automation_id"
            ),
            ["status_automation_id"],
            unique=False,
        )

    with op.batch_alter_table(
        "project_task_status_link", schema=None
    ) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_project_task_status_link_project_id"),
            ["project_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_project_task_status_link_task_status_id"),
            ["task_status_id"],
            unique=False,
        )

    with op.batch_alter_table(
        "project_task_type_link", schema=None
    ) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_project_task_type_link_project_id"),
            ["project_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_project_task_type_link_task_type_id"),
            ["task_type_id"],
            unique=False,
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table(
        "project_task_type_link", schema=None
    ) as batch_op:
        batch_op.drop_index(
            batch_op.f("ix_project_task_type_link_task_type_id")
        )
        batch_op.drop_index(batch_op.f("ix_project_task_type_link_project_id"))

    with op.batch_alter_table(
        "project_task_status_link", schema=None
    ) as batch_op:
        batch_op.drop_index(
            batch_op.f("ix_project_task_status_link_task_status_id")
        )
        batch_op.drop_index(
            batch_op.f("ix_project_task_status_link_project_id")
        )

    with op.batch_alter_table(
        "project_status_automation_link", schema=None
    ) as batch_op:
        batch_op.drop_index(
            batch_op.f(
                "ix_project_status_automation_link_status_automation_id"
            )
        )
        batch_op.drop_index(
            batch_op.f("ix_project_status_automation_link_project_id")
        )

    with op.batch_alter_table("project_person_link", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_project_person_link_project_id"))
        batch_op.drop_index(batch_op.f("ix_project_person_link_person_id"))

    with op.batch_alter_table("comment_preview_link", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_comment_preview_link_preview_file"))
        batch_op.drop_index(batch_op.f("ix_comment_preview_link_comment"))

    # ### end Alembic commands ###



from . import routes
from flask import Blueprint
from zou.app.utils.api import configure_api_from_blueprint


def init_plugin(app, manifest):
    """
    Init the plugin.
    """
    app.logger.info("Loading plugin...")
    routes_tuples = [(f"/hello-world", routes.HelloWorld)]

    blueprint = Blueprint(manifest["id"], manifest["id"])
    configure_api_from_blueprint(blueprint, routes_tuples)
    app.register_blueprint(blueprint, url_prefix=f"/{manifest['id']}")


def pre_install(manifest):
    """
    Pre install the plugin.
    """


def post_install(manifest):
    """
    Post install the plugin.
    """


def pre_uninstall(manifest):
    """
    Pre uninstall the plugin.
    """


def post_uninstall(manifest):
    """
    Post uninstall the plugin.
    """



"""Add new constraint for TimeSpent

Revision ID: 16328eae4b5f
Revises: 17ef8f7be758
Create Date: 2024-03-06 05:14:09.534846

"""

from alembic import op
import sqlalchemy as sa
from sqlalchemy.orm.session import Session
from zou.migrations.utils.base import BaseMixin
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy_utils import UUIDType

# revision identifiers, used by Alembic.
revision = "16328eae4b5f"
down_revision = "17ef8f7be758"
branch_labels = None
depends_on = None

base = declarative_base()


class TimeSpent(base, BaseMixin):
    """
    Describes the time spent by someone on a task.
    """

    __tablename__ = "time_spent"

    duration = sa.Column(sa.Float, nullable=False)
    date = sa.Column(sa.Date, nullable=False)

    task_id = sa.Column(
        UUIDType(binary=False), sa.ForeignKey("task.id"), index=True
    )
    person_id = sa.Column(
        UUIDType(binary=False), sa.ForeignKey("person.id"), index=True
    )

    __table_args__ = (
        sa.UniqueConstraint(
            "person_id", "task_id", "date", name="time_spent_uc"
        ),
    )


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    with op.batch_alter_table("time_spent", schema=None) as batch_op:
        session = Session(bind=op.get_bind())
        session.query(TimeSpent).where(TimeSpent.duration <= 0).delete()
        session.commit()
        batch_op.create_check_constraint(
            "check_duration_positive", "duration > 0"
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("time_spent", schema=None) as batch_op:
        batch_op.drop_constraint("check_duration_positive", type_="check")

    # ### end Alembic commands ###



"""drop column TaskType.for_shots

Revision ID: 0cf5e0e035fa
Revises: f874ad5e898a
Create Date: 2022-06-10 13:34:48.830690

"""

from alembic import op
import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy_utils import UUIDType
from sqlalchemy.ext.declarative import declarative_base
from zou.migrations.utils.base import BaseMixin


# revision identifiers, used by Alembic.
revision = "0cf5e0e035fa"
down_revision = "f874ad5e898a"
branch_labels = None
depends_on = None


class TaskType(declarative_base(), BaseMixin):
    """
    Categorize tasks in domain areas: modeling, animation, etc.
    """

    __tablename__ = "task_type"
    name = sa.Column(sa.String(40), nullable=False)
    short_name = sa.Column(sa.String(20))
    color = sa.Column(sa.String(7), default="#FFFFFF")
    priority = sa.Column(sa.Integer, default=1)
    for_entity = sa.Column(sa.String(30), default="Asset")
    allow_timelog = sa.Column(sa.Boolean, default=True)
    shotgun_id = sa.Column(sa.Integer, index=True)
    for_shots = sa.Column(sa.Boolean, default=False)
    department_id = sa.Column(
        UUIDType(binary=False), sa.ForeignKey("department.id")
    )

    __table_args__ = (
        sa.UniqueConstraint(
            "name", "for_entity", "department_id", name="task_type_uc"
        ),
    )


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    bind = op.get_bind()
    session = orm.Session(bind=bind)
    task_types = session.query(TaskType).all()
    for task_type in task_types:
        if task_type.for_shots and task_type.for_entity != "Shot":
            task_type.for_entity = "Shot"
    session.commit()
    op.drop_column("task_type", "for_shots")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "task_type",
        sa.Column(
            "for_shots", sa.BOOLEAN(), autoincrement=False, nullable=True
        ),
    )
    # ### end Alembic commands ###



"""Add some indexes

Revision ID: 16df47d76c64
Revises: 0596674df51d
Create Date: 2023-07-24 22:44:26.233788

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "16df47d76c64"
down_revision = "0596674df51d"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("comment", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_comment_person_id"), ["person_id"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_comment_task_status_id"),
            ["task_status_id"],
            unique=False,
        )

    with op.batch_alter_table("department_link", schema=None) as batch_op:
        batch_op.alter_column(
            "person_id", existing_type=sa.UUID(), nullable=False
        )
        batch_op.alter_column(
            "department_id", existing_type=sa.UUID(), nullable=False
        )

    with op.batch_alter_table("output_file", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_output_file_entity_id"), ["entity_id"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_output_file_file_status_id"),
            ["file_status_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_output_file_person_id"), ["person_id"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_output_file_source_file_id"),
            ["source_file_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_output_file_temporal_entity_id"),
            ["temporal_entity_id"],
            unique=False,
        )

    with op.batch_alter_table("search_filter", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_search_filter_person_id"),
            ["person_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_search_filter_project_id"),
            ["project_id"],
            unique=False,
        )

    with op.batch_alter_table("task_type", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_task_type_department_id"),
            ["department_id"],
            unique=False,
        )

    with op.batch_alter_table("working_file", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_working_file_person_id"),
            ["person_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_working_file_software_id"),
            ["software_id"],
            unique=False,
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("working_file", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_working_file_software_id"))
        batch_op.drop_index(batch_op.f("ix_working_file_person_id"))

    with op.batch_alter_table("task_type", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_task_type_department_id"))

    with op.batch_alter_table("search_filter", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_search_filter_project_id"))
        batch_op.drop_index(batch_op.f("ix_search_filter_person_id"))

    with op.batch_alter_table("output_file", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_output_file_temporal_entity_id"))
        batch_op.drop_index(batch_op.f("ix_output_file_source_file_id"))
        batch_op.drop_index(batch_op.f("ix_output_file_person_id"))
        batch_op.drop_index(batch_op.f("ix_output_file_file_status_id"))
        batch_op.drop_index(batch_op.f("ix_output_file_entity_id"))

    with op.batch_alter_table("department_link", schema=None) as batch_op:
        batch_op.alter_column(
            "department_id", existing_type=sa.UUID(), nullable=True
        )
        batch_op.alter_column(
            "person_id", existing_type=sa.UUID(), nullable=True
        )

    with op.batch_alter_table("comment", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_comment_task_status_id"))
        batch_op.drop_index(batch_op.f("ix_comment_person_id"))

    # ### end Alembic commands ###



import subprocess


class ShellCommandFailed(Exception):
    pass


def run_command(args):
    try:
        sp = subprocess.Popen(
            args, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        out, err = sp.communicate()
        if err:
            raise ShellCommandFailed(err)
        else:
            return out
    except Exception as e:
        raise ShellCommandFailed(str(e))



from sqlalchemy_utils import UUIDType

from zou.app import db
from zou.app.models.serializer import SerializerMixin
from zou.app.models.base import BaseMixin


class Subscription(db.Model, BaseMixin, SerializerMixin):
    """
    Allow to subscribe to an entity
    """

    person_id = db.Column(
        UUIDType(binary=False),
        db.ForeignKey("person.id"),
        nullable=False,
        index=True,
    )
    task_id = db.Column(
        UUIDType(binary=False), db.ForeignKey("task.id"), index=True
    )

    entity_id = db.Column(
        UUIDType(binary=False), db.ForeignKey("entity.id"), index=True
    )  # Deprecated
    task_type_id = db.Column(
        UUIDType(binary=False), db.ForeignKey("task_type.id"), index=True
    )  # Deprecated

    __table_args__ = (
        db.UniqueConstraint(
            "person_id", "task_id", name="subscription_task_uc"
        ),
        db.UniqueConstraint(
            "person_id",
            "task_type_id",
            "entity_id",
            name="subscription_entity_uc",
        ),
    )



from zou.app import db
from zou.app.models.serializer import SerializerMixin
from zou.app.models.base import BaseMixin
from sqlalchemy_utils import EmailType, URLType


class Plugin(db.Model, BaseMixin, SerializerMixin):
    """
    Describe a plugin.
    """

    plugin_id = db.Column(
        db.String(80), unique=True, nullable=False, index=True
    )
    name = db.Column(db.String(80), nullable=False, index=True)
    description = db.Column(db.Text())
    version = db.Column(db.String(50), nullable=False)
    maintainer_name = db.Column(db.String(200), nullable=False)
    maintainer_email = db.Column(EmailType)
    website = db.Column(URLType)
    license = db.Column(db.String(80), nullable=False)



"""add for entity field to playlists

Revision ID: 7417c8eb70d8
Revises: cf3d365de164
Create Date: 2020-03-15 17:49:32.365732

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "7417c8eb70d8"
down_revision = "8739ae9fa28b"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "playlist",
        sa.Column("for_entity", sa.String(length=10), nullable=True),
    )
    op.create_index(
        op.f("ix_playlist_for_entity"),
        "playlist",
        ["for_entity"],
        unique=False,
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_playlist_for_entity"), table_name="playlist")
    op.drop_column("playlist", "for_entity")
    # ### end Alembic commands ###



from gevent import monkey

monkey.patch_all()

from flask import jsonify
from flask_jwt_extended import (
    get_jwt_identity,
    jwt_required,
    verify_jwt_in_request,
)
from flask_socketio import SocketIO, disconnect, join_room, leave_room, emit

from zou.app import config, app
from zou.app.utils.redis import get_redis_url

server_stats = {"nb_connections": 0}
rooms_data = {}

redis_url = get_redis_url(config.KV_EVENTS_DB_INDEX)
socketio = SocketIO(
    logger=True, cors_allowed_origins=[], cors_credentials=False
)
socketio.init_app(app, message_queue=redis_url, async_mode="gevent")


def _get_empty_room(current_frame=0):
    return {
        "playlist_id": None,
        "user_id": None,
        "local_id": None,
        "people": [],
        "is_playing": False,
        "current_entity_id": None,
        "current_entity_index": None,
        "current_preview_file_id": None,
        "current_preview_file_index": None,
        "current_frame": current_frame,
        "is_repeating": None,
        "is_annotations_displayed": False,
        "is_zoom_enabled": False,
        "is_waveform_displayed": False,
        "is_laser_mode": None,
        "handle_in": None,
        "handle_out": None,
        "speed": None,
        "comparing": {
            "enable": False,
            "task_type": None,
            "revision": None,
            "mode": "sidebyside",
            "comparison_preview_index": 0,
        },
    }


def _get_room_from_data(data):
    room_id = data["playlist_id"]
    return rooms_data.get(room_id, _get_empty_room()), room_id


def _leave_room(room_id, user_id):
    room = rooms_data.get(room_id, _get_empty_room())
    room["people"] = list(set(room["people"]) - {user_id})
    if len(room["people"]) > 0:
        rooms_data[room_id] = room
    elif room_id in rooms_data:
        del rooms_data[room_id]
    _emit_people_updated(room_id, room["people"])
    return room


def _emit_people_updated(room_id, people):
    event_data = {
        "people": people,
        "playlist_id": room_id,
        "id": room_id,
    }
    emit("preview-room:room-people-updated", event_data, room=room_id)
    return event_data


def _update_room_playing_status(data, room):
    room["playlist_id"] = data.get("playlist_id", False)
    room["user_id"] = data.get("user_id", False)
    room["local_id"] = data.get("local_id", False)
    room["is_playing"] = data.get("is_playing", False)
    room["is_repeating"] = data.get("is_repeating", False)
    room["is_laser_mode"] = data.get("is_laser_mode", False)
    room["is_annotations_displayed"] = data.get(
        "is_annotations_displayed", False
    )
    room["is_zoom_enabled"] = data.get("is_zoom_enabled", False)
    room["is_waveform_displayed"] = data.get("is_waveform_displayed", False)
    room["current_entity_id"] = data.get("current_entity_id", None)
    room["current_entity_index"] = data.get("current_entity_index", None)
    room["current_preview_file_id"] = data.get("current_preview_file_id", None)
    room["current_preview_file_index"] = data.get(
        "current_preview_file_index", None
    )
    room["handle_in"] = data.get("handle_in", None)
    room["handle_out"] = data.get("handle_out", None)
    if "current_frame" in data:
        room["current_frame"] = data["current_frame"]
    if "comparing" in data:
        room["comparing"] = data["comparing"]
    if "speed" in data:
        room["speed"] = data["speed"]
    return room


@app.route("/", methods=["GET"])
def index():
    return jsonify({"name": "%s Event stream" % config.APP_NAME})


@app.route("/stats", methods=["GET"])
def stats():
    return jsonify(server_stats)


@socketio.on("connect", namespace="/events")
def connected(_):
    try:
        verify_jwt_in_request()
        server_stats["nb_connections"] += 1
        app.logger.info("New websocket client connected")
    except Exception:
        app.logger.info("New websocket client failed to connect")
        disconnect()
        return False


@socketio.on("disconnect", namespace="/events")
def disconnected(_):
    try:
        verify_jwt_in_request()
        user_id = get_jwt_identity()
        # Needed to be able to clear empty rooms
        tmp_rooms_data = dict(rooms_data)
        for room_id in tmp_rooms_data:
            _leave_room(room_id, user_id)
            leave_room(room_id, user_id)
        server_stats["nb_connections"] -= 1
        app.logger.info("Websocket client disconnected")
    except Exception:
        pass


@socketio.on_error("/events")
def on_error(error):
    server_stats["nb_connections"] -= 1
    if server_stats["nb_connections"] < 0:
        server_stats["nb_connections"] = 0
    app.logger.error(error)


@app.route("/rooms", methods=["GET", "POST"])
@jwt_required()
def rooms():
    return jsonify({"name": "%s Review rooms" % config.APP_NAME})


@socketio.on("preview-room:open-playlist", namespace="/events")
@jwt_required()
def on_open_playlist(data):
    """
    when a person opens the playlist page he immediately enters the
    websocket room. This way he can see in live which people are in the
    review room. The user still has to explicitly enter the review room
    to actually be in sync with the other users.
    """
    room, room_id = _get_room_from_data(data)
    rooms_data[room_id] = room
    # Connect to the socketio room but dont add the user to the data of
    # the room.
    join_room(room_id)
    _emit_people_updated(room_id, room["people"])


@socketio.on("preview-room:close-playlist", namespace="/events")
@jwt_required()
def on_close_playlist(data):
    """
    when a person closes the playlist page he immediately leaves the
    websocket room.
    """
    room, room_id = _get_room_from_data(data)
    # Leave only the socketio room but dont remove the user from the data of
    # the room. This operation must be done via a leave event.
    leave_room(room_id)


@socketio.on("preview-room:join", namespace="/events")
@jwt_required()
def on_join(data):
    """
    When a person joins the review room, we notify all its members that a
    new person is added to the room.
    """
    user_id = get_jwt_identity()
    room, room_id = _get_room_from_data(data)
    if len(room["people"]) == 0:
        _update_room_playing_status(data, room)
    room["people"] = list(set(room["people"] + [user_id]))
    room["playlist_id"] = room_id
    rooms_data[room_id] = room
    _emit_people_updated(room_id, room["people"])
    emit("preview-room:room-updated", room, room=room_id)


@socketio.on("preview-room:leave", namespace="/events")
@jwt_required()
def on_leave(data):
    user_id = get_jwt_identity()
    room_id = data.get("playlist_id", "")
    _leave_room(room_id, user_id)


@socketio.on("preview-room:room-updated", namespace="/events")
@jwt_required()
def on_room_updated(data, only_newcomer=False):
    room, room_id = _get_room_from_data(data)
    rooms_data[room_id] = _update_room_playing_status(data, room)
    event_data = {"only_newcomer": only_newcomer, **rooms_data[room_id]}
    emit("preview-room:room-updated", event_data, room=room_id)


@socketio.on("preview-room:add-annotation", namespace="/events")
@jwt_required()
def on_add_annotation(data):
    room_id = data["playlist_id"]
    emit("preview-room:add-annotation", data, room=room_id)


@socketio.on("preview-room:remove-annotation", namespace="/events")
@jwt_required()
def on_remove_annotation(data):
    room_id = data["playlist_id"]
    emit("preview-room:remove-annotation", data, room=room_id)


@socketio.on("preview-room:update-annotation", namespace="/events")
@jwt_required()
def on_update_annotation(data):
    room_id = data["playlist_id"]
    emit("preview-room:update-annotation", data, room=room_id)


@socketio.on("preview-room:change-version", namespace="/events")
@jwt_required()
def on_change_version(data):
    room_id = data["playlist_id"]
    emit("preview-room:change-version", data, room=room_id)


@socketio.on("preview-room:panzoom-changed", namespace="/events")
@jwt_required()
def on_change_version(data):
    room_id = data["playlist_id"]
    emit("preview-room:panzoom-changed", data, room=room_id)


if __name__ == "__main__":
    socketio.run(
        app,
        debug=False,
        host=config.EVENT_STREAM_HOST,
        port=config.EVENT_STREAM_PORT,
    )



"""Add start and end dates to projects

Revision ID: 003be8a91001
Revises: 590aa1ffe731
Create Date: 2019-07-05 09:22:47.857496

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "003be8a91001"
down_revision = "590aa1ffe731"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("project", sa.Column("end_date", sa.Date(), nullable=True))
    op.add_column("project", sa.Column("start_date", sa.Date(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("project", "start_date")
    op.drop_column("project", "end_date")
    # ### end Alembic commands ###



"""add acks to comments

Revision ID: 3476e147e632
Revises: 6aa446ee4072
Create Date: 2020-04-10 00:50:31.704488

"""

from alembic import op
import sqlalchemy as sa
import sqlalchemy_utils
import sqlalchemy_utils
import uuid

# revision identifiers, used by Alembic.
revision = "3476e147e632"
down_revision = "6aa446ee4072"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "comment_acknoledgments",
        sa.Column(
            "comment",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=False,
        ),
        sa.Column(
            "person",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["comment"],
            ["comment.id"],
        ),
        sa.ForeignKeyConstraint(
            ["person"],
            ["person.id"],
        ),
        sa.PrimaryKeyConstraint("comment", "person"),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("comment_acknoledgments")
    # ### end Alembic commands ###



from sqlalchemy_utils import UUIDType
from sqlalchemy_utils import ChoiceType

from zou.app.models.person import POSITION_TYPES, SENIORITY_TYPES

from zou.app import db
from zou.app.models.serializer import SerializerMixin
from zou.app.models.base import BaseMixin


class SalaryScale(db.Model, BaseMixin, SerializerMixin):
    """
    Model to represent a salary scale tied to a department.
    """

    department_id = db.Column(
        UUIDType(binary=False),
        db.ForeignKey("department.id"),
        index=True,
        nullable=False,
    )

    position = db.Column(ChoiceType(POSITION_TYPES), default="artist")
    seniority = db.Column(ChoiceType(SENIORITY_TYPES), default="mid")
    salary = db.Column(db.Integer, nullable=False, default=0)

    def present(self):
        return self.serialize()



from sqlalchemy_utils import UUIDType

from zou.app import db
from zou.app.models.serializer import SerializerMixin
from zou.app.models.base import BaseMixin
from sqlalchemy.sql import expression


class SearchFilterGroup(db.Model, BaseMixin, SerializerMixin):
    """
    Groups are used to store search filters into sections.
    """

    list_type = db.Column(db.String(80), nullable=False, index=True)
    entity_type = db.Column(db.String(80))
    name = db.Column(db.String(200), nullable=False, default="")
    color = db.Column(db.String(8), nullable=False, default="")
    is_shared = db.Column(
        db.Boolean,
        server_default=expression.false(),
        default=False,
        nullable=False,
    )

    person_id = db.Column(UUIDType(binary=False), db.ForeignKey("person.id"))
    project_id = db.Column(UUIDType(binary=False), db.ForeignKey("project.id"))
    department_id = db.Column(
        UUIDType(binary=False), db.ForeignKey("department.id")
    )



"""add entity status field

Revision ID: 2baede80b111
Revises: 956659992419
Create Date: 2023-01-24 19:09:55.648115

"""

from alembic import op
import sqlalchemy as sa
import sqlalchemy_utils

# revision identifiers, used by Alembic.

revision = "2baede80b111"
down_revision = "956659992419"
branch_labels = None
depends_on = None

ENTITY_STATUSES = [
    ("standby", "Stand By"),
    ("running", "Running"),
    ("complete", "Complete"),
    ("canceled", "Canceled"),
]


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "entity",
        sa.Column(
            "status",
            sqlalchemy_utils.types.choice.ChoiceType(ENTITY_STATUSES),
            nullable=True,
            default="running",
        ),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("entity", "status")
    # ### end Alembic commands ###



from sqlalchemy_utils import UUIDType

from zou.app import db
from zou.app.models.serializer import SerializerMixin
from zou.app.models.base import BaseMixin
from zou.app.models.task_type import TaskType


class TaskTypeAssetTypeLink(db.Model):
    asset_type_id = db.Column(
        UUIDType(binary=False),
        db.ForeignKey("entity_type.id"),
        index=True,
        primary_key=True,
    )
    task_type_id = db.Column(
        UUIDType(binary=False),
        db.ForeignKey("task_type.id"),
        index=True,
        primary_key=True,
    )

    __table_args__ = (
        db.UniqueConstraint(
            "asset_type_id",
            "task_type_id",
            name="task_type_asset_type_link_uc",
        ),
    )


class EntityType(db.Model, BaseMixin, SerializerMixin):
    """
    Type of entities. It can describe either an asset type, or tell if target
    entity is a shot, sequence, episode or layout scene.
    """

    name = db.Column(db.String(30), unique=True, nullable=False, index=True)
    short_name = db.Column(db.String(20))
    description = db.Column(db.Text())
    task_types = db.relationship(
        TaskType,
        secondary=TaskTypeAssetTypeLink.__table__,
        lazy="joined",
    )
    archived = db.Column(db.Boolean(), default=False)

    @classmethod
    def create_from_import(cls, data):
        is_update = False
        task_types_ids = None
        if "type" in data:
            del data["type"]
        if "task_types" in data:
            task_types_ids = data.pop("task_types", None)
        entity_type = cls.get(data["id"])
        if entity_type is None:
            entity_type = cls.create(**data)
            is_update = False
        else:
            entity_type.update(data)
            is_update = True
        if task_types_ids is not None:
            entity_type.task_types = []
            for task_type_id in task_types_ids:
                task_type = TaskType.get(task_type_id)
                if task_type is not None:
                    entity_type.task_types.append(task_type)
            entity_type.save()
        return (entity_type, is_update)



"""empty message

Revision ID: f0567e8d0c62
Revises: a23682ccc1f1
Create Date: 2018-04-26 17:24:47.093547

"""

from alembic import op
import sqlalchemy as sa
import sqlalchemy_utils
import uuid

# revision identifiers, used by Alembic.
revision = "f0567e8d0c62"
down_revision = "a23682ccc1f1"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "entity",
        sa.Column(
            "source_id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=True,
        ),
    )
    op.create_index(
        op.f("ix_entity_source_id"), "entity", ["source_id"], unique=False
    )
    op.create_foreign_key(None, "entity", "entity", ["source_id"], ["id"])
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(None, "entity", type_="foreignkey")
    op.drop_index(op.f("ix_entity_source_id"), table_name="entity")
    op.drop_column("entity", "source_id")
    # ### end Alembic commands ###



from gevent import monkey

monkey.patch_all()

from zou.app import app, config
from flask_socketio import SocketIO
import logging


FORMAT = "%(message)s"
logging.basicConfig(level=logging.INFO, format=FORMAT)
socketio = SocketIO(app, cors_allowed_origins=[], cors_credentials=False)

if __name__ == "__main__":
    print(
        f"The Kitsu API server is listening on http://{config.DEBUG_HOST}:{config.DEBUG_PORT}"
    )
    socketio.run(
        app, host=config.DEBUG_HOST, port=config.DEBUG_PORT, debug=True
    )



"""empty message

Revision ID: a519c710877c
Revises: 4f2398ebcd49
Create Date: 2019-02-25 13:39:29.662284

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "a519c710877c"
down_revision = "4f2398ebcd49"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "entity", sa.Column("code", sa.String(length=160), nullable=True)
    )
    op.add_column(
        "project", sa.Column("code", sa.String(length=80), nullable=True)
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("project", "code")
    op.drop_column("entity", "code")
    # ### end Alembic commands ###



"""Introduce Plugin table

Revision ID: 4aab1f84ad72
Revises: d25118cddcaa
Create Date: 2025-04-14 15:18:29.346896

"""

from alembic import op
import sqlalchemy as sa
import sqlalchemy_utils
import sqlalchemy_utils
import uuid

# revision identifiers, used by Alembic.
revision = "4aab1f84ad72"
down_revision = "d25118cddcaa"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "plugin",
        sa.Column("plugin_id", sa.String(length=80), nullable=False),
        sa.Column("name", sa.String(length=80), nullable=False),
        sa.Column("description", sa.Text(), nullable=True),
        sa.Column("version", sa.String(length=50), nullable=False),
        sa.Column("maintainer_name", sa.String(length=200), nullable=False),
        sa.Column(
            "maintainer_email",
            sqlalchemy_utils.types.email.EmailType(length=255),
            nullable=True,
        ),
        sa.Column(
            "website", sqlalchemy_utils.types.url.URLType(), nullable=True
        ),
        sa.Column("license", sa.String(length=80), nullable=False),
        sa.Column(
            "id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=False,
        ),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("updated_at", sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    with op.batch_alter_table("plugin", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_plugin_name"), ["name"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_plugin_plugin_id"), ["plugin_id"], unique=True
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("plugin", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_plugin_plugin_id"))
        batch_op.drop_index(batch_op.f("ix_plugin_name"))

    op.drop_table("plugin")
    # ### end Alembic commands ###



"""empty message

Revision ID: 818f7bda2528
Revises: bf1347acdee2
Create Date: 2018-04-10 22:45:19.462757

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "818f7bda2528"
down_revision = "bf1347acdee2"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "preview_file", sa.Column("path", sa.String(length=400), nullable=True)
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("preview_file", "path")
    # ### end Alembic commands ###



from sqlalchemy_utils import (
    UUIDType,
    EmailType,
    LocaleType,
    TimezoneType,
    ChoiceType,
)
from sqlalchemy import Index
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.dialects.postgresql import JSONB

from pytz import timezone as pytz_timezone
from babel import Locale

from zou.app.models.serializer import SerializerMixin
from zou.app.models.department import Department
from zou.app.models.base import BaseMixin
from zou.app import config, db


TWO_FACTOR_AUTHENTICATION_TYPES = [
    ("totp", "TOTP"),
    ("email_otp", "Email OTP"),
    ("fido", "FIDO"),
]

CONTRACT_TYPES = [
    ("open-ended", "Open-ended"),
    ("fixed-term", "Fixed-term"),
    ("short-term", "Short-term"),
    ("freelance", "Freelance"),
    ("apprentice", "Apprentice"),
    ("internship", "Internship"),
]

ROLE_TYPES = [
    ("user", "Artist"),
    ("admin", "Studio Manager"),
    ("supervisor", "Supervisor"),
    ("manager", "Production Manager"),
    ("client", "Client"),
    ("vendor", "Vendor"),
]

POSITION_TYPES = [
    ("supervisor", "Supervisor"),
    ("lead", "Lead"),
    ("artist", "Artist"),
]

SENIORITY_TYPES = [
    ("senior", "Senior"),
    ("mid", "Mid"),
    ("junior", "Junior"),
]


class DepartmentLink(db.Model):
    __tablename__ = "department_link"
    person_id = db.Column(
        UUIDType(binary=False),
        db.ForeignKey("person.id"),
        primary_key=True,
        index=True,
    )
    department_id = db.Column(
        UUIDType(binary=False),
        db.ForeignKey("department.id"),
        primary_key=True,
        index=True,
    )

    __table_args__ = (
        db.UniqueConstraint(
            "person_id",
            "department_id",
            name="department_link_uc",
        ),
    )


class Person(db.Model, BaseMixin, SerializerMixin):
    """
    Describe a member of the studio (and an API user).
    """

    first_name = db.Column(db.String(80), nullable=False)
    last_name = db.Column(db.String(80), nullable=False)
    email = db.Column(EmailType)
    phone = db.Column(db.String(30))
    contract_type = db.Column(
        ChoiceType(CONTRACT_TYPES), default="open-ended", nullable=False
    )

    active = db.Column(db.Boolean(), default=True)
    archived = db.Column(db.Boolean(), default=False)
    last_presence = db.Column(db.Date())

    password = db.Column(db.LargeBinary(60))
    desktop_login = db.Column(db.String(80))
    login_failed_attemps = db.Column(db.Integer, default=0)
    last_login_failed = db.Column(db.DateTime())
    totp_enabled = db.Column(db.Boolean(), default=False)
    totp_secret = db.Column(db.String(32), default=None)
    email_otp_enabled = db.Column(db.Boolean(), default=False)
    email_otp_secret = db.Column(db.String(32), default=None)
    fido_enabled = db.Column(db.Boolean(), default=False)
    fido_credentials = db.Column(db.ARRAY(JSONB))
    otp_recovery_codes = db.Column(db.ARRAY(db.LargeBinary(60)))
    preferred_two_factor_authentication = db.Column(
        ChoiceType(TWO_FACTOR_AUTHENTICATION_TYPES)
    )

    shotgun_id = db.Column(db.Integer, unique=True)
    timezone = db.Column(
        TimezoneType(backend="pytz"),
        default=pytz_timezone(config.DEFAULT_TIMEZONE),
    )
    locale = db.Column(LocaleType, default=Locale(config.DEFAULT_LOCALE))
    data = db.Column(JSONB)
    role = db.Column(ChoiceType(ROLE_TYPES), default="user", nullable=False)
    position = db.Column(ChoiceType(POSITION_TYPES), default="artist")
    seniority = db.Column(ChoiceType(SENIORITY_TYPES), default="mid")
    daily_salary = db.Column(db.Integer, default=0)

    has_avatar = db.Column(db.Boolean(), default=False)

    notifications_enabled = db.Column(db.Boolean(), default=False)
    notifications_slack_enabled = db.Column(db.Boolean(), default=False)
    notifications_slack_userid = db.Column(db.String(60), default="")
    notifications_mattermost_enabled = db.Column(db.Boolean(), default=False)
    notifications_mattermost_userid = db.Column(db.String(60), default="")
    notifications_discord_enabled = db.Column(db.Boolean(), default=False)
    notifications_discord_userid = db.Column(db.String(60), default="")

    is_bot = db.Column(db.Boolean(), default=False, nullable=False)
    jti = db.Column(db.String(60), nullable=True, unique=True)
    expiration_date = db.Column(db.Date(), nullable=True)

    departments = db.relationship(
        "Department", secondary="department_link", lazy="joined"
    )
    studio_id = db.Column(
        UUIDType(binary=False), db.ForeignKey("studio.id"), index=True
    )

    is_generated_from_ldap = db.Column(db.Boolean(), default=False)
    ldap_uid = db.Column(db.String(60), unique=True, default=None)

    __table_args__ = (
        Index(
            "only_one_email_by_person",
            email,
            is_bot,
            unique=True,
            postgresql_where=is_bot.isnot(True),
        ),
    )

    def __repr__(self):
        return f"<Person {self.full_name}>"

    @hybrid_property
    def full_name(self):
        if self.first_name and self.last_name:
            return f"{self.first_name} {self.last_name}"
        else:
            return f"{self.first_name}{self.last_name}"

    @full_name.expression
    def full_name(cls):
        if cls.first_name and cls.last_name:
            return cls.first_name + " " + cls.last_name
        else:
            return cls.first_name + cls.last_name

    def fido_devices(self):
        if self.fido_credentials is None:
            return []
        else:
            return [
                credential["device_name"]
                for credential in self.fido_credentials
            ]

    def serialize(
        self, obj_type="Person", relations=False, milliseconds=False
    ):
        data = SerializerMixin.serialize(
            self, obj_type, relations=relations, milliseconds=milliseconds
        )
        data["fido_devices"] = self.fido_devices()
        return data

    def serialize_safe(self, relations=False, milliseconds=False):
        data = self.serialize(relations=relations, milliseconds=milliseconds)
        del data["password"]
        del data["totp_secret"]
        del data["email_otp_secret"]
        del data["otp_recovery_codes"]
        del data["fido_credentials"]
        del data["jti"]
        return data

    def present_minimal(self, relations=False, milliseconds=False):
        data = SerializerMixin.serialize(
            self, "Person", relations=relations, milliseconds=milliseconds
        )
        return {
            "id": data["id"],
            "type": data["type"],
            "first_name": data["first_name"],
            "last_name": data["last_name"],
            "full_name": self.full_name,
            "has_avatar": data["has_avatar"],
            "active": data["active"],
            "departments": data.get("departments", []),
            "studio_id": data["studio_id"],
            "role": data["role"],
            "desktop_login": data["desktop_login"],
            "is_bot": data["is_bot"],
        }

    def set_departments(self, department_ids):
        self.departments = []
        for department_id in department_ids:
            department = Department.get(department_id)
            if department is not None:
                self.departments.append(department)
        self.save()

    @classmethod
    def create_from_import(cls, person):
        del person["type"]
        del person["full_name"]
        is_update = False
        previous_person = cls.get(person["id"])

        if "password" in person and person["password"] is not None:
            person["password"] = person["password"].encode()

        department_ids = None
        if "departments" in person:
            department_ids = person.pop("departments", None)

        if previous_person is None:
            previous_person = cls.create(**person)
        else:
            is_update = True
            previous_person.update(person)

        if department_ids is not None:
            previous_person.set_departments(department_ids)

        return (previous_person, is_update)



"""add new column MetadataDescriptor.data_type + migrate data

Revision ID: 9e5b3a9b0cee
Revises: 8fbd40afbe5f
Create Date: 2023-03-22 13:26:27.580633

"""

from alembic import op
import sqlalchemy as sa
import sqlalchemy_utils
from sqlalchemy import orm
from sqlalchemy.ext.declarative import declarative_base
from zou.migrations.utils.base import BaseMixin
from sqlalchemy.dialects import postgresql


# revision identifiers, used by Alembic.
revision = "9e5b3a9b0cee"
down_revision = "8fbd40afbe5f"
branch_labels = None
depends_on = None


METADATA_DESCRIPTOR_TYPES = [
    ("string", "String"),
    ("number", "Number"),
    ("list", "List"),
    ("boolean", "Boolean"),
    ("checklist", "Checklist"),
]

base = declarative_base()


class Department(base, BaseMixin):
    """
    Studio department like modeling, animation, etc.
    """

    __tablename__ = "department"
    name = sa.Column(sa.String(80), unique=True, nullable=False)
    color = sa.Column(sa.String(7), nullable=False)


department_metadata_descriptor_link = sa.Table(
    "department_metadata_descriptor_link",
    base.metadata,
    sa.Column(
        "metadata_descriptor_id",
        sqlalchemy_utils.UUIDType(binary=False),
        sa.ForeignKey("metadata_descriptor.id"),
    ),
    sa.Column(
        "department_id",
        sqlalchemy_utils.UUIDType(binary=False),
        sa.ForeignKey("department.id"),
    ),
)


class MetadataDescriptor(base, BaseMixin):
    """
    This models allow to identify which metadata are available for a given
    project and a given entity type.
    """

    __tablename__ = "metadata_descriptor"

    project_id = sa.Column(
        sqlalchemy_utils.UUIDType(binary=False),
        sa.ForeignKey("project.id"),
        nullable=False,
        index=True,
    )
    entity_type = sa.Column(sa.String(60), nullable=False, index=True)
    name = sa.Column(sa.String(120), nullable=False)
    data_type = sa.Column(
        sqlalchemy_utils.types.choice.ChoiceType(METADATA_DESCRIPTOR_TYPES)
    )
    field_name = sa.Column(sa.String(120), nullable=False)
    choices = sa.Column(postgresql.JSONB(astext_type=sa.Text()))
    for_client = sa.Column(sa.Boolean(), default=False, index=True)
    departments = orm.relationship(
        "Department", secondary=department_metadata_descriptor_link
    )

    __table_args__ = (
        sa.UniqueConstraint(
            "project_id", "entity_type", "name", name="metadata_descriptor_uc"
        ),
        sa.UniqueConstraint(
            "project_id",
            "entity_type",
            "field_name",
            name="metadata_descriptor_uc2",
        ),
    )


PROJECT_STYLES = [
    ("2d", "2D Animation"),
    ("3d", "3D Animation"),
    ("2d3d", "2D/3D Animation"),
    ("vfx", "VFX"),
    ("stop-motion", "Stop Motion"),
    ("motion-design", "Motion Design"),
    ("archviz", "Archviz"),
    ("commercial", "Commercial"),
    ("catalog", "Catalog"),
]


class Project(base, BaseMixin):
    __tablename__ = "project"

    name = sa.Column(sa.String(80), nullable=False, unique=True, index=True)
    code = sa.Column(sa.String(80))
    description = sa.Column(sa.Text())
    shotgun_id = sa.Column(sa.Integer)
    file_tree = sa.Column(postgresql.JSONB(astext_type=sa.Text()))
    data = sa.Column(postgresql.JSONB(astext_type=sa.Text()))
    has_avatar = sa.Column(sa.Boolean(), default=False)
    fps = sa.Column(sa.String(10))
    ratio = sa.Column(sa.String(10))
    resolution = sa.Column(sa.String(12))
    production_type = sa.Column(sa.String(20), default="short")
    production_style = sa.Column(
        sqlalchemy_utils.types.choice.ChoiceType(PROJECT_STYLES)
    )
    start_date = sa.Column(sa.Date())
    end_date = sa.Column(sa.Date())
    man_days = sa.Column(sa.Integer)
    nb_episodes = sa.Column(sa.Integer, default=0)
    episode_span = sa.Column(sa.Integer, default=0)
    max_retakes = sa.Column(sa.Integer, default=0)
    is_clients_isolated = sa.Column(sa.Boolean(), default=False)

    project_status_id = sa.Column(
        sqlalchemy_utils.UUIDType(binary=False),
        sa.ForeignKey("project_status.id"),
        index=True,
    )


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("metadata_descriptor", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "data_type",
                sqlalchemy_utils.types.choice.ChoiceType(
                    METADATA_DESCRIPTOR_TYPES
                ),
                nullable=True,
            )
        )
    # ### end Alembic commands ###

    bind = op.get_bind()
    session = orm.Session(bind=bind)
    metadata_descriptors = session.query(MetadataDescriptor).all()
    for metadata_descriptor in metadata_descriptors:
        if len(metadata_descriptor.choices) > 1 and all(
            isinstance(s, str) for s in metadata_descriptor.choices
        ):
            if all(
                s.startswith("[ ]") or s.startswith("[x]")
                for s in metadata_descriptor.choices
            ):
                metadata_descriptor.data_type = "checklist"
            else:
                metadata_descriptor.data_type = "list"
        else:
            metadata_descriptor.data_type = "string"
    session.commit()


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("metadata_descriptor", schema=None) as batch_op:
        batch_op.drop_column("data_type")

    # ### end Alembic commands ###



from zou.app.blueprints.crud.base import BaseModelsResource, BaseModelResource

from zou.app.models.salary_scale import Budget


class BudgetsResource(BaseModelsResource):
    def __init__(self):
        BaseModelsResource.__init__(self, Budget)


class BudgetResource(BaseModelResource):
    protected_fields = [
        "id",
        "created_at",
        "updated_at",
        "project_id",
        "revision",
    ]

    def __init__(self):
        BaseModelResource.__init__(self, Budget)



from werkzeug.exceptions import NotFound, Forbidden, BadRequest, Unauthorized


class EpisodeNotFoundException(NotFound):
    pass


class SequenceNotFoundException(NotFound):
    pass


class ShotNotFoundException(NotFound):
    pass


class ConceptNotFoundException(NotFound):
    pass


class SceneNotFoundException(NotFound):
    pass


class AssetNotFoundException(NotFound):
    pass


class AssetInstanceNotFoundException(NotFound):
    pass


class AssetTypeNotFoundException(NotFound):
    pass


class AttachmentFileNotFoundException(NotFound):
    pass


class TaskNotFoundException(NotFound):
    pass


class PreviewBackgroundFileNotFoundException(NotFound):
    pass


class DepartmentNotFoundException(NotFound):
    pass


class StudioNotFoundException(NotFound):
    pass


class TaskStatusNotFoundException(NotFound):
    pass


class TaskTypeNotFoundException(NotFound):
    pass


class PersonInProtectedAccounts(Forbidden):
    pass


class PersonNotFoundException(NotFound):
    pass


class TimeSpentNotFoundException(NotFound):
    pass


class ProjectNotFoundException(NotFound):
    pass


class WorkingFileNotFoundException(NotFound):
    pass


class OutputFileNotFoundException(NotFound):
    pass


class SoftwareNotFoundException(NotFound):
    pass


class OutputTypeNotFoundException(NotFound):
    pass


class PreviewFileNotFoundException(NotFound):
    pass


class PreviewFileReuploadNotAllowedException(BadRequest):
    pass


class CommentNotFoundException(NotFound):
    pass


class NewsNotFoundException(NotFound):
    pass


class EntityNotFoundException(NotFound):
    pass


class EntityLinkNotFoundException(NotFound):
    pass


class EntityTypeNotFoundException(NotFound):
    pass


class BuildJobNotFoundException(NotFound):
    pass


class PlaylistNotFoundException(NotFound):
    pass


class SearchFilterNotFoundException(NotFound):
    pass


class SearchFilterGroupNotFoundException(NotFound):
    pass


class NotificationNotFoundException(NotFound):
    pass


class SubscriptionNotFoundException(NotFound):
    pass


class MetadataDescriptorNotFoundException(NotFound):
    pass


class BudgetNotFoundException(NotFound):
    pass


class BudgetEntryNotFoundException(NotFound):
    pass


class MalformedFileTreeException(Exception):
    pass


class WrongFileTreeFileException(Exception):
    pass


class WrongPathFormatException(Exception):
    pass


class NoOutputFileException(Exception):
    pass


class NoAuthStrategyConfigured(Exception):
    pass


class WrongUserException(Exception):
    pass


class WrongPasswordException(Exception):
    pass


class MissingOTPException(Exception):
    def __init__(
        self,
        preferred_two_factor_authentication,
        two_factor_authentication_enabled,
    ):
        self.preferred_two_factor_authentication = (
            preferred_two_factor_authentication
        )
        self.two_factor_authentication_enabled = (
            two_factor_authentication_enabled
        )


class WrongOTPException(Exception):
    pass


class TOTPAlreadyEnabledException(Exception):
    pass


class TOTPNotEnabledException(Exception):
    pass


class TwoFactorAuthenticationNotEnabledException(Exception):
    pass


class FIDONoPreregistrationException(Exception):
    pass


class FIDOServerException(Exception):
    pass


class FIDONotEnabledException(Exception):
    pass


class EmailOTPAlreadyEnabledException(Exception):
    pass


class EmailOTPNotEnabledException(Exception):
    pass


class NoTwoFactorAuthenticationEnabled(Exception):
    pass


class TooMuchLoginFailedAttemps(Exception):
    pass


class UserCantConnectDueToNoFallback(Exception):
    pass


class UnactiveUserException(Unauthorized):
    pass


class WrongDateFormatException(Exception):
    pass


class EntryAlreadyExistsException(Exception):
    pass


class WrongParameterException(Exception):
    def __init__(self, message, dict=None):
        super().__init__(message)
        self.dict = dict


class WrongIdFormatException(Exception):
    pass


class ModelWithRelationsDeletionException(Exception):
    pass


class EditNotFoundException(NotFound):
    pass


class StatusAutomationNotFoundException(NotFound):
    pass


class WrongTaskTypeForEntityException(Exception):
    pass


class IsUserLimitReachedException(Exception):
    pass



"""add max retake parameter

Revision ID: 5a291251823c
Revises: 4095103c7d01
Create Date: 2022-06-29 10:56:13.556495

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "5a291251823c"
down_revision = "4095103c7d01"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "project", sa.Column("max_retakes", sa.Integer(), nullable=True)
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("project", "max_retakes")
    # ### end Alembic commands ###



from flask import Blueprint
from zou.app.utils.api import configure_api_from_blueprint

from zou.app.blueprints.breakdown.resources import (
    EpisodesCastingResource,
    ProjectEntityLinkResource,
    ProjectEntityLinksResource,
    ShotAssetInstancesResource,
    RemoveShotAssetInstanceResource,
    SceneAssetInstancesResource,
    SceneCameraInstancesResource,
    CastingResource,
    AssetTypeCastingResource,
    SequenceCastingResource,
    EpisodeSequenceAllCastingResource,
    SequenceAllCastingResource,
)


routes = [
    (
        "/data/projects/<project_id>/entities/<entity_id>/casting",
        CastingResource,
    ),
    (
        "/data/projects/<project_id>/asset-types/<asset_type_id>/casting",
        AssetTypeCastingResource,
    ),
    (
        "/data/projects/<project_id>/episodes/casting",
        EpisodesCastingResource,
    ),
    (
        "/data/projects/<project_id>/sequences/<sequence_id>/casting",
        SequenceCastingResource,
    ),
    (
        "/data/projects/<project_id>/episodes/<episode_id>/sequences/all/casting",
        EpisodeSequenceAllCastingResource,
    ),
    (
        "/data/projects/<project_id>/sequences/all/casting",
        SequenceAllCastingResource,
    ),
    ("/data/projects/<project_id>/entity-links", ProjectEntityLinksResource),
    (
        "/data/projects/<project_id>/entity-links/<entity_link_id>",
        ProjectEntityLinkResource,
    ),
    ("/data/scenes/<scene_id>/asset-instances", SceneAssetInstancesResource),
    ("/data/scenes/<scene_id>/camera-instances", SceneCameraInstancesResource),
    ("/data/shots/<shot_id>/asset-instances", ShotAssetInstancesResource),
    (
        "/data/shots/<shot_id>/asset-instances/<asset_instance_id>",
        RemoveShotAssetInstanceResource,
    ),
]


blueprint = Blueprint("breakdown", "breakdown")
api = configure_api_from_blueprint(blueprint, routes)



from flask import Blueprint
from zou.app.utils.api import configure_api_from_blueprint

from zou.app.blueprints.concepts.resources import (
    ConceptResource,
    AllConceptsResource,
    ConceptsAndTasksResource,
    ConceptPreviewsResource,
    ConceptTaskTypesResource,
    ConceptTasksResource,
    ProjectConceptsResource,
)

routes = [
    ("/data/concepts", AllConceptsResource),
    ("/data/concepts/with-tasks", ConceptsAndTasksResource),
    ("/data/concepts/<concept_id>", ConceptResource),
    ("/data/concepts/<concept_id>/task-types", ConceptTaskTypesResource),
    ("/data/concepts/<concept_id>/tasks", ConceptTasksResource),
    ("/data/concepts/<concept_id>/preview-files", ConceptPreviewsResource),
    ("/data/projects/<project_id>/concepts", ProjectConceptsResource),
]


blueprint = Blueprint("concepts", "concepts")
api = configure_api_from_blueprint(blueprint, routes)



"""Empty revision due to wrong revision creation

Revision ID: 5ab9d7a75887
Revises: d80267806131
Create Date: 2022-06-06 22:33:26.331874

"""

# revision identifiers, used by Alembic.
revision = "5ab9d7a75887"
down_revision = "d80267806131"
branch_labels = None
depends_on = None


def upgrade():
    pass


def downgrade():
    pass



from zou.app.blueprints.crud.base import BaseModelsResource, BaseModelResource

from zou.app.models.salary_scale import BudgetEntry


class BudgetEntriesResource(BaseModelsResource):

    def __init__(self):
        BaseModelsResource.__init__(self, BudgetEntry)


class BudgetEntryResource(BaseModelResource):

    def __init__(self):
        BaseModelResource.__init__(self, BudgetEntry)



"""For DepartmentLink index person_id / department_id

Revision ID: 539a3a00c417
Revises: 9d3bb33c6fc6
Create Date: 2025-01-14 12:19:52.699322

"""

from alembic import op


# revision identifiers, used by Alembic.
revision = "539a3a00c417"
down_revision = "9d3bb33c6fc6"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("department_link", schema=None) as batch_op:
        batch_op.create_unique_constraint(
            "department_link_uc", ["person_id", "department_id"]
        )
        batch_op.create_index(
            batch_op.f("ix_department_link_department_id"),
            ["department_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_department_link_person_id"),
            ["person_id"],
            unique=False,
        )
        batch_op.create_primary_key(
            "department_link_pkey", ["person_id", "department_id"]
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("department_link", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_department_link_person_id"))
        batch_op.drop_index(batch_op.f("ix_department_link_department_id"))
        batch_op.drop_constraint("department_link_uc", type_="unique")
        batch_op.drop_constraint("department_link_pkey", type_="primary")

    # ### end Alembic commands ###



import zipfile
import semver
import shutil
from pathlib import Path

from zou.app import config, db
from zou.app.models.plugin import Plugin
from zou.app.utils.plugins import PluginManifest


def install_plugin(path, force=False):
    """
    Install a plugin.
    """
    path = Path(path)
    if not path.exists():
        raise FileNotFoundError(f"Plugin path '{path}' does not exist.")

    manifest = PluginManifest.from_plugin_path(path)
    plugin = Plugin.query.filter_by(plugin_id=manifest.id).one_or_none()

    try:
        already_installed = False
        if plugin:
            current = semver.Version.parse(plugin.version)
            new = semver.Version.parse(str(manifest.version))
            if not force and new <= current:
                raise ValueError(
                    f"Plugin version {new} is not newer than {current}."
                )
            plugin.update_no_commit(manifest.to_model_dict())
            already_installed = True
        else:
            plugin = Plugin.create_no_commit(**manifest.to_model_dict())

        install_plugin_files(manifest.id, path, already_installed)
    except Exception:
        uninstall_plugin_files(manifest.id)
        db.session.rollback()
        db.session.remove()
        raise

    Plugin.commit()
    return plugin.serialize()


def install_plugin_files(plugin_id, path, already_installed=False):
    """
    Install plugin files.
    """
    path = Path(path)
    plugin_path = Path(config.PLUGIN_FOLDER) / plugin_id
    if already_installed and plugin_path.exists():
        shutil.rmtree(plugin_path)

    plugin_path.mkdir(parents=True, exist_ok=True)

    if path.is_dir():
        shutil.copytree(path, plugin_path, dirs_exist_ok=True)
    elif zipfile.is_zipfile(path):
        shutil.unpack_archive(path, plugin_path, format="zip")
    else:
        raise ValueError(
            f"Plugin path '{path}' is not a valid zip file or a directory."
        )

    return plugin_path


def uninstall_plugin_files(plugin_id):
    """
    Uninstall plugin files.
    """
    plugin_path = Path(config.PLUGIN_FOLDER) / plugin_id
    if plugin_path.exists():
        shutil.rmtree(plugin_path)
        return True
    return False


def uninstall_plugin(plugin_id):
    """
    Uninstall a plugin.
    """
    installed = uninstall_plugin_files(plugin_id)
    plugin = Plugin.query.filter_by(plugin_id=plugin_id).one_or_none()
    if plugin is not None:
        installed = True
        plugin.delete()

    if not installed:
        raise ValueError(f"Plugin '{plugin_id}' is not installed.")
    return True


def create_plugin_skeleton(
    path,
    id,
    name,
    description=None,
    version=None,
    maintainer=None,
    website=None,
    license=None,
    force=False,
):
    plugin_template_path = (
        Path(__file__).parent.parent.parent / "plugin_template"
    )
    plugin_path = Path(path) / id

    if plugin_path.exists():
        if force:
            shutil.rmtree(plugin_path)
        else:
            raise FileExistsError(
                f"Plugin '{id}' already exists in {plugin_path}."
            )

    shutil.copytree(plugin_template_path, plugin_path)

    manifest = PluginManifest.from_file(plugin_path / "manifest.toml")

    manifest.id = id
    manifest.name = name
    if description:
        manifest.description = description
    if version:
        manifest.version = version
    if maintainer:
        manifest.maintainer = maintainer
    if website:
        manifest.website = website
    if license:
        manifest.license = license

    manifest.validate()
    manifest.write_to_path(plugin_path)

    return plugin_path


def create_plugin_package(path, output_path, force=False):
    """
    Create a plugin package.
    """
    path = Path(path)
    if not path.exists():
        raise FileNotFoundError(f"Plugin path '{path}' does not exist.")

    manifest = PluginManifest.from_plugin_path(path)

    output_path = Path(output_path)
    if not output_path.suffix == ".zip":
        output_path /= f"{manifest.id}-{manifest.version}.zip"
    if output_path.exists():
        if force:
            output_path.unlink()
        else:
            raise FileExistsError(
                f"Output path '{output_path}' already exists."
            )

    output_path = shutil.make_archive(
        output_path.with_suffix(""),
        "zip",
        path,
    )
    return output_path



import itertools
from operator import itemgetter

from zou.app.models.file_status import FileStatus
from zou.app import app

from zou.app.models.entity import Entity
from zou.app.models.output_file import OutputFile
from zou.app.models.output_type import OutputType
from zou.app.models.preview_file import PreviewFile
from zou.app.models.preview_background_file import PreviewBackgroundFile
from zou.app.models.software import Software
from zou.app.models.task import Task
from zou.app.models.working_file import WorkingFile


from zou.app.services import entities_service
from zou.app.services.base_service import (
    get_instance,
    get_or_create_instance_by_name,
)

from zou.app.services.exception import (
    WorkingFileNotFoundException,
    OutputFileNotFoundException,
    OutputTypeNotFoundException,
    PreviewFileNotFoundException,
    SoftwareNotFoundException,
    NoOutputFileException,
    EntryAlreadyExistsException,
    PreviewBackgroundFileNotFoundException,
)

from zou.app.utils import cache, fields, events, query as query_utils

from sqlalchemy import desc, func
from sqlalchemy.exc import StatementError, IntegrityError
from sqlalchemy.sql.expression import and_


def clear_preview_file_cache(preview_file_id):
    cache.cache.delete_memoized(get_preview_file, preview_file_id)


def clear_output_file_cache(output_file_id):
    cache.cache.delete_memoized(get_output_file, output_file_id)


def clear_working_file_cache(working_file_id):
    cache.cache.delete_memoized(get_working_file, working_file_id)


def clear_output_type_cache(output_type_id):
    cache.cache.delete_memoized(get_output_type, output_type_id)


def clear_software_cache(software_id):
    cache.cache.delete_memoized(get_software, software_id)


def clear_preview_background_file_cache(preview_background_file_id):
    cache.cache.delete_memoized(
        get_preview_background_file, preview_background_file_id
    )
    cache.cache.delete_memoized(get_preview_background_files)


@cache.memoize_function(240)
def get_default_status():
    """
    Return default file status to set on a file when it is created.
    """
    default_status = FileStatus.get_by(name=app.config["DEFAULT_FILE_STATUS"])
    if default_status is None:
        default_status = FileStatus(
            name=app.config["DEFAULT_FILE_STATUS"], color="#FFFFFF"
        )
        default_status.save()
    return default_status.serialize()


def get_working_file_raw(working_file_id):
    """
    Return given working file as active record.
    """
    return get_instance(
        WorkingFile, working_file_id, WorkingFileNotFoundException
    )


@cache.memoize_function(240)
def get_working_file(working_file_id):
    """
    Return given working file as dict.
    """
    return get_working_file_raw(working_file_id).serialize(relations=True)


def get_output_file_raw(output_file_id):
    """
    Return given output file as active record.
    """
    return get_instance(
        OutputFile, output_file_id, OutputFileNotFoundException
    )


@cache.memoize_function(240)
def get_output_file(output_file_id):
    """
    Return given output file as a dict.
    """
    return get_output_file_raw(output_file_id).serialize()


def get_software_raw(software_id):
    """
    Return given software as active record.
    """
    return get_instance(Software, software_id, SoftwareNotFoundException)


@cache.memoize_function(240)
def get_software(software_id):
    """
    Return given software as dict.
    """
    return get_software_raw(software_id).serialize()


def get_output_type_raw(output_type_id):
    """
    Return given output type as active record.
    """
    return get_instance(
        OutputType, output_type_id, OutputTypeNotFoundException
    )


@cache.memoize_function(240)
def get_output_type(output_type_id):
    """
    Return given output type as dict.
    """
    return get_output_type_raw(output_type_id).serialize()


def get_or_create_output_type(name, short_name=""):
    return get_or_create_instance_by_name(
        OutputType, name=name, short_name=short_name
    )


def get_or_create_software(name, short_name, file_extension):
    return get_or_create_instance_by_name(
        Software,
        name=name,
        short_name=short_name,
        file_extension=file_extension,
    )


def get_last_working_files_for_task(task_id):
    """
    Get last revisions for given task grouped by file name.
    """
    query = WorkingFile.query.with_entities(
        WorkingFile.name,
        WorkingFile.task_id,
        func.max(WorkingFile.revision).label("MAX"),
    ).group_by(
        WorkingFile.name,
        WorkingFile.task_id,
    )

    query = query.filter(WorkingFile.task_id == task_id)
    statement = query.subquery()

    query = WorkingFile.query.join(
        statement,
        and_(
            WorkingFile.task_id == statement.c.task_id,
            WorkingFile.name == statement.c.name,
            WorkingFile.revision == statement.c.MAX,
        ),
    )

    # query
    working_files = fields.serialize_models(query.all())

    # group by name
    working_files_by_name = {
        k: list(v)[0]
        for k, v in itertools.groupby(working_files, key=itemgetter("name"))
    }

    return working_files_by_name


def get_next_working_revision(task_id, name):
    """
    Get next working file revision for given task and name.
    """
    working_files = (
        WorkingFile.query.filter_by(task_id=task_id, name=name)
        .order_by(desc(WorkingFile.revision))
        .all()
    )
    if len(working_files) > 0:
        revision = working_files[0].revision + 1
    else:
        revision = 1
    return revision


def create_new_working_revision(
    task_id,
    person_id,
    software_id,
    name="main",
    path="",
    comment="",
    revision=0,
):
    """
    Create a new working file revision for given task. An author (user) and
    a software are required.
    """
    task = Task.get(task_id)
    if revision == 0:
        revision = get_next_working_revision(task_id, name)

    try:
        working_file = WorkingFile.create(
            comment=comment,
            name=name,
            revision=revision,
            path=path,
            task_id=task.id,
            software_id=software_id,
            entity_id=task.entity_id,
            person_id=person_id,
        )
        events.emit(
            "working-file:new",
            {"working_file_id": working_file.id},
            project_id=str(task.project_id),
        )
    except IntegrityError:
        raise EntryAlreadyExistsException

    return working_file.serialize()


def create_new_output_revision(
    entity_id,
    working_file_id,
    output_type_id,
    person_id,
    task_type_id,
    revision=0,
    representation="",
    name="main",
    comment="",
    extension="",
    nb_elements=1,
    asset_instance_id=None,
    temporal_entity_id=None,
    file_status_id=None,
):
    """
    Create a new ouput file for given entity. Output type, task type, author
    and source file are required.

    The revision is set as next revision available but it can be forced.
    An extension and a name can be set too.

    An asset instance can be given too. In that case, the output file is
    linked to the asset instance.

    The `temporal_entity_id` concerns only asset instance output files. It is
    here to describe if the output is generated in the context of a shot or in
    the context of a scene.
    """

    if revision < 1:
        try:
            output_file = get_last_output_revision(
                entity_id,
                output_type_id,
                task_type_id,
                name=name,
                asset_instance_id=asset_instance_id,
                temporal_entity_id=temporal_entity_id,
            )

            revision = output_file["revision"] + 1
        except NoOutputFileException:
            revision = 1

    file_status_id = file_status_id or get_default_status()["id"]

    try:
        output_file = OutputFile.get_by(
            name=name,
            entity_id=entity_id,
            asset_instance_id=asset_instance_id,
            output_type_id=output_type_id,
            task_type_id=task_type_id,
            temporal_entity_id=temporal_entity_id,
            representation=representation,
            revision=revision,
        )

        if output_file is None:
            output_file = OutputFile.create(
                name=name,
                comment=comment,
                extension=extension,
                representation=representation,
                revision=revision,
                entity_id=entity_id,
                asset_instance_id=asset_instance_id,
                person_id=person_id,
                source_file_id=working_file_id,
                output_type_id=output_type_id,
                file_status_id=file_status_id,
                task_type_id=task_type_id,
                nb_elements=nb_elements,
                temporal_entity_id=temporal_entity_id,
            )
            entity = entities_service.get_entity(entity_id)
            events.emit(
                "output-file:new",
                {"output_file_id": output_file.id},
                project_id=entity["project_id"],
            )
        else:
            raise EntryAlreadyExistsException

    except IntegrityError:
        raise EntryAlreadyExistsException

    return output_file.serialize()


def get_working_files_for_task(task_id):
    """
    Retrieve all working files for a given task ordered by revision from
    biggest to smallest revision.
    """
    working_files = (
        WorkingFile.query.filter_by(task_id=task_id)
        .filter(WorkingFile.revision >= 0)
        .order_by(desc(WorkingFile.revision))
        .all()
    )
    return fields.serialize_models(working_files)


def get_working_files_for_entity(
    entity_id, task_id=None, name=None, relations=False
):
    """
    Retrieve all working files for a given entity and specified parameters
    ordered by revision from biggest to smallest revision.
    """
    query = WorkingFile.query.filter_by(entity_id=entity_id)

    if task_id:
        query = query.filter(WorkingFile.task_id == task_id)
    if name:
        query = query.filter(WorkingFile.name == name)

    query = query.filter(WorkingFile.revision >= 0).order_by(
        desc(WorkingFile.revision)
    )

    working_files = query.all()
    return fields.serialize_models(working_files, relations=relations)


def get_next_working_file_revision(task_id, name):
    """
    Get next working file revision available for given task and given name.
    """
    last_working_files = get_last_working_files_for_task(task_id)
    working_file = last_working_files.get(name, None)
    if working_file is not None:
        revision = working_file["revision"] + 1
    else:
        revision = 1
    return revision


def get_next_output_file_revision(
    entity_id,
    output_type_id,
    task_type_id,
    name="main",
    asset_instance_id=None,
    temporal_entity_id=None,
):
    """
    Get next output file revision available for given entity, output type, task
    type and name.
    Asset instance can be set, in that case, it looks for next revision
    available for it instead of entity.
    """
    try:
        last_output = get_last_output_revision(
            output_type_id=output_type_id,
            task_type_id=task_type_id,
            entity_id=entity_id,
            name=name,
            asset_instance_id=asset_instance_id,
            temporal_entity_id=temporal_entity_id,
        )
        return last_output["revision"] + 1
    except NoOutputFileException:
        return 1


def get_last_output_revision(
    entity_id,
    output_type_id,
    task_type_id,
    name="main",
    asset_instance_id=None,
    temporal_entity_id=None,
):
    """
    Get output with highest revision created for given entity, output type, task
    type and name.
    If an asset instance is given, it will look for last output file for this
    instance instead of given entity.
    """
    query = (
        OutputFile.query.filter_by(
            output_type_id=output_type_id, task_type_id=task_type_id, name=name
        )
        .filter(OutputFile.revision > 0)
        .order_by(desc(OutputFile.revision))
    )

    if asset_instance_id is None:
        query = query.filter(OutputFile.entity_id == entity_id)
    else:
        query = query.filter(
            OutputFile.asset_instance_id == asset_instance_id,
            OutputFile.temporal_entity_id == temporal_entity_id,
        )

    output_files = query.all()

    if len(output_files) == 0:
        raise NoOutputFileException()

    return output_files[0].serialize()


def get_output_files_for_project(
    project_id,
    task_type_id=None,
    output_type_id=None,
    name=None,
    representation=None,
    file_status_id=None,
):
    """
    Return output files for given project ordered by creation date.
    """
    query = OutputFile.query

    if task_type_id:
        query = query.filter(OutputFile.task_type_id == task_type_id)
    if output_type_id:
        query = query.filter(OutputFile.output_type_id == output_type_id)
    if name:
        query = query.filter(OutputFile.name == name)
    if representation:
        query = query.filter(OutputFile.representation == representation)
    if file_status_id:
        query = query.filter(OutputFile.file_status_id == file_status_id)

    output_files = (
        query.filter(OutputFile.asset_instance_id == None)
        .filter(OutputFile.temporal_entity_id == None)
        .filter(OutputFile.revision >= 0)
        .filter(Entity.project_id == project_id)
        .join(Entity, OutputFile.entity_id == Entity.id)
        .order_by(desc(OutputFile.created_at))
        .all()
    )
    return fields.serialize_models(output_files)


def get_output_files_for_entity(
    entity_id,
    task_type_id=None,
    output_type_id=None,
    name=None,
    representation=None,
    file_status_id=None,
):
    """
    Return output files for given entity ordered by revision.
    """
    query = OutputFile.query.filter_by(entity_id=entity_id)

    if task_type_id:
        query = query.filter(OutputFile.task_type_id == task_type_id)
    if output_type_id:
        query = query.filter(OutputFile.output_type_id == output_type_id)
    if name:
        query = query.filter(OutputFile.name == name)
    if representation:
        query = query.filter(OutputFile.representation == representation)
    if file_status_id:
        query = query.filter(OutputFile.file_status_id == file_status_id)

    query = query.filter(OutputFile.asset_instance_id == None).filter(
        OutputFile.temporal_entity_id == None
    )

    output_files = (
        query.filter(OutputFile.revision >= 0)
        .order_by(desc(OutputFile.revision))
        .all()
    )
    return fields.serialize_models(output_files)


def get_output_files_for_instance(
    asset_instance_id,
    temporal_entity_id,
    task_type_id=None,
    output_type_id=None,
    name=None,
    representation=None,
    file_status_id=None,
):
    """
    Return output files for given instance ordered by revision.
    """
    query = OutputFile.query.filter_by(asset_instance_id=asset_instance_id)

    if temporal_entity_id:
        query = query.filter(
            OutputFile.temporal_entity_id == temporal_entity_id
        )
    if task_type_id:
        query = query.filter(OutputFile.task_type_id == task_type_id)
    if output_type_id:
        query = query.filter(OutputFile.output_type_id == output_type_id)
    if name:
        query = query.filter(OutputFile.name == name)
    if representation:
        query = query.filter(OutputFile.representation == representation)
    if file_status_id:
        query = query.filter(OutputFile.file_status_id == file_status_id)

    output_files = (
        query.filter(OutputFile.revision >= 0)
        .order_by(desc(OutputFile.revision))
        .all()
    )
    return fields.serialize_models(output_files)


def get_last_output_files_for_entity(
    entity_id,
    task_type_id=None,
    output_type_id=None,
    name=None,
    representation=None,
    file_status_id=None,
):
    """
    Get last output files for given parameters.

    We use a subquery to get maximum revision and then filter with given
    params.
    """
    # Query maximum revision for each possible arguments.
    # Only group and filter by file status if specified. Otherwise this could
    # lead to different groups of files sharing the same history but with
    # different status.
    # This could be very misleading when the user would want to get the last.
    if file_status_id:
        query = OutputFile.query.with_entities(
            OutputFile.task_type_id,
            OutputFile.output_type_id,
            OutputFile.name,
            OutputFile.representation,
            OutputFile.file_status_id,
            func.max(OutputFile.revision).label("MAX"),
        ).group_by(
            OutputFile.task_type_id,
            OutputFile.output_type_id,
            OutputFile.name,
            OutputFile.representation,
            OutputFile.file_status_id,
        )
    else:
        query = OutputFile.query.with_entities(
            OutputFile.task_type_id,
            OutputFile.output_type_id,
            OutputFile.name,
            OutputFile.representation,
            func.max(OutputFile.revision).label("MAX"),
        ).group_by(
            OutputFile.task_type_id,
            OutputFile.output_type_id,
            OutputFile.name,
            OutputFile.representation,
        )

    query = query.filter(OutputFile.entity_id == entity_id).filter(
        OutputFile.asset_instance_id == None
    )
    if file_status_id:
        query = query.filter(OutputFile.file_status_id == file_status_id)
    if task_type_id:
        query = query.filter(OutputFile.task_type_id == task_type_id)
    if output_type_id:
        query = query.filter(OutputFile.output_type_id == output_type_id)
    if name:
        query = query.filter(OutputFile.name == name)
    if representation:
        query = query.filter(OutputFile.representation == representation)
    statement = query.subquery()

    # Create a join query to retrieve maximum revision and filter by
    # specified arguments
    query = OutputFile.query.join(
        statement,
        and_(
            OutputFile.entity_id == entity_id,
            OutputFile.asset_instance_id == None,
            OutputFile.task_type_id == statement.c.task_type_id,
            OutputFile.output_type_id == statement.c.output_type_id,
            OutputFile.name == statement.c.name,
            OutputFile.representation == statement.c.representation,
            OutputFile.revision == statement.c.MAX,
        ),
    )

    # query
    output_files = query.all()
    return fields.serialize_models(output_files)


def get_last_output_files_for_instance(
    asset_instance_id,
    temporal_entity_id,
    task_type_id=None,
    output_type_id=None,
    name=None,
    representation=None,
    file_status_id=None,
):
    """
    Get last output files for given entity grouped by output type and name.
    """
    # Query maximum revision for each possible arguments
    # Only group and filter by file status if specified. Otherwise this could
    # lead to different groups of files sharing the same history but with
    # different status.
    # This could be very misleading when the user would want to get the last.
    if file_status_id:
        query = OutputFile.query.with_entities(
            OutputFile.asset_instance_id,
            OutputFile.temporal_entity_id,
            OutputFile.task_type_id,
            OutputFile.output_type_id,
            OutputFile.name,
            OutputFile.representation,
            OutputFile.file_status_id,
            func.max(OutputFile.revision).label("MAX"),
        ).group_by(
            OutputFile.asset_instance_id,
            OutputFile.temporal_entity_id,
            OutputFile.task_type_id,
            OutputFile.output_type_id,
            OutputFile.name,
            OutputFile.representation,
            OutputFile.file_status_id,
        )
    else:
        query = OutputFile.query.with_entities(
            OutputFile.asset_instance_id,
            OutputFile.temporal_entity_id,
            OutputFile.task_type_id,
            OutputFile.output_type_id,
            OutputFile.name,
            OutputFile.representation,
            func.max(OutputFile.revision).label("MAX"),
        ).group_by(
            OutputFile.asset_instance_id,
            OutputFile.temporal_entity_id,
            OutputFile.task_type_id,
            OutputFile.output_type_id,
            OutputFile.name,
            OutputFile.representation,
        )
    query = query.filter(OutputFile.asset_instance_id == asset_instance_id)
    query = query.filter(OutputFile.temporal_entity_id == temporal_entity_id)
    if file_status_id:
        query = query.filter(OutputFile.file_status_id == file_status_id)
    if task_type_id:
        query = query.filter(OutputFile.task_type_id == task_type_id)
    if output_type_id:
        query = query.filter(OutputFile.output_type_id == output_type_id)
    if name:
        query = query.filter(OutputFile.name == name)
    if representation:
        query = query.filter(OutputFile.representation == representation)
    statement = query.subquery()

    # Create a join query to retrieve maximum revision
    query = OutputFile.query.join(
        statement,
        and_(
            OutputFile.asset_instance_id == statement.c.asset_instance_id,
            OutputFile.temporal_entity_id == statement.c.temporal_entity_id,
            OutputFile.task_type_id == statement.c.task_type_id,
            OutputFile.output_type_id == statement.c.output_type_id,
            OutputFile.name == statement.c.name,
            OutputFile.representation == statement.c.representation,
            OutputFile.revision == statement.c.MAX,
        ),
    )

    output_files = query.all()
    return fields.serialize_models(output_files)


def get_preview_file_raw(preview_file_id):
    """
    Get preview file as active record.
    """
    try:
        preview_file = PreviewFile.get(preview_file_id)
    except StatementError:
        raise PreviewFileNotFoundException()

    if preview_file is None:
        raise PreviewFileNotFoundException()

    return preview_file


@cache.memoize_function(240)
def get_preview_file(preview_file_id):
    """
    Get preview file as dict.
    """
    preview_file = get_preview_file_raw(preview_file_id)
    return preview_file.serialize()


def get_preview_files_for_task(task_id):
    """
    Get all preview files for given task.
    """
    previews = PreviewFile.query.filter_by(task_id=task_id).order_by(
        PreviewFile.revision.desc()
    )
    return fields.serialize_models(previews)


def create_preview_file_raw(
    name,
    revision,
    task_id,
    person_id,
    source="webgui",
    extension="mp4",
    position=1,
):
    return PreviewFile.create(
        name=name,
        revision=revision,
        source=source,
        task_id=task_id,
        person_id=person_id,
        extension=extension,
        position=position,
        status="processing",
    )


def create_preview_file(
    name,
    revision,
    task_id,
    person_id,
    source="webgui",
    extension="mp4",
    position=1,
):
    return create_preview_file_raw(
        name, revision, task_id, person_id, source, extension, position
    ).serialize()


def update_working_file(working_file_id, data):
    working_file = get_working_file_raw(working_file_id)
    working_file.update(data)
    clear_working_file_cache(working_file_id)
    return working_file.serialize()


def update_output_file(output_file_id, data):
    output_file = get_output_file_raw(output_file_id)
    output_file.update(data)
    clear_output_file_cache(output_file_id)
    return output_file.serialize()


def get_output_types_for_entity(entity_id):
    """
    Get output types from all output files created for given entity.
    """
    output_types = (
        OutputType.query.join(OutputFile)
        .filter(OutputFile.entity_id == entity_id)
        .order_by(OutputType.name)
        .all()
    )
    return OutputType.serialize_list(output_types)


def get_output_types_for_instance(asset_instance_id, temporal_entity_id=None):
    """
    Get output types from all output files created for given instance.
    """
    output_types = (
        OutputType.query.join(OutputFile)
        .filter(
            OutputFile.asset_instance_id == asset_instance_id,
            OutputFile.temporal_entity_id == temporal_entity_id,
        )
        .order_by(OutputType.name)
        .all()
    )
    return OutputType.serialize_list(output_types)


def get_output_files_for_output_type_and_entity(
    entity_id, output_type_id, representation=None
):
    """
    Get output files created for given entity and output type.
    """
    query = (
        OutputFile.query.filter(OutputFile.entity_id == entity_id)
        .filter(OutputFile.asset_instance_id == None)  # noqa Entity only
        .filter(OutputFile.output_type_id == output_type_id)
        .order_by(desc(OutputFile.revision))
    )

    if representation is not None:
        query = query.filter(OutputFile.representation == representation)

    output_files = query.all()
    return OutputFile.serialize_list(output_files)


def get_output_files_for_output_type_and_asset_instance(
    asset_instance_id, temporal_entity_id, output_type_id, representation=None
):
    """
    Get output files created for given asset instance and output type.
    """
    query = OutputFile.query.filter(
        OutputFile.asset_instance_id == asset_instance_id,
        OutputFile.output_type_id == output_type_id,
        OutputFile.temporal_entity_id == temporal_entity_id,
    ).order_by(desc(OutputFile.revision))

    if representation is not None:
        query = query.filter(OutputFile.representation == representation)

    output_files = query.all()
    return OutputFile.serialize_list(output_files)


def remove_preview_file(preview_file_id):
    preview_file = get_preview_file_raw(preview_file_id)
    preview_file.delete()
    task = Task.get(preview_file.task_id)
    events.emit(
        "preview-file:delete",
        {"preview_file_id": preview_file_id},
        project_id=str(task.project_id),
    )
    return preview_file.serialize()


def get_preview_files_for_project(project_id, page=-1):
    """
    Return all preview files for given project.
    """
    query = (
        PreviewFile.query.join(Task)
        .filter(Task.project_id == project_id)
        .order_by(desc(PreviewFile.updated_at))
    )
    return query_utils.get_paginated_results(query, page)


def get_preview_background_file_raw(preview_background_file_id):
    """
    Get preview background file matching given id as an active record.
    """
    try:
        preview_background = PreviewBackgroundFile.get(
            preview_background_file_id
        )
    except StatementError:
        raise PreviewBackgroundFileNotFoundException

    if preview_background is None:
        raise PreviewBackgroundFileNotFoundException

    return preview_background


@cache.memoize_function(1200)
def get_preview_background_file(preview_background_file_id):
    """
    Get preview background file matching given id as a dictionary.
    """
    return get_preview_background_file_raw(
        preview_background_file_id
    ).serialize()


@cache.memoize_function(120)
def get_preview_background_files():
    """
    Get all preview backgrounds files.
    """
    return fields.serialize_models(PreviewBackgroundFile.get_all())


def reset_default_preview_background_files(preview_background_file_id):
    """
    Set all preview background files as is_default=False except the one matching given id.
    """
    PreviewBackgroundFile.query.filter(
        PreviewBackgroundFile.id != preview_background_file_id,
        PreviewBackgroundFile.is_default == True,
    ).update({"is_default": False})
    PreviewBackgroundFile.commit()


def update_preview_background_file(preview_background_file_id, data):
    preview_background_file = get_preview_background_file_raw(
        preview_background_file_id
    )
    preview_background_file.update(data)
    clear_preview_background_file_cache(preview_background_file_id)
    return preview_background_file.serialize()



from slugify import slugify
from zou.app.blueprints.source.csv.base import BaseCsvProjectImportResource

from zou.app.services import (
    assets_service,
    projects_service,
    shots_service,
    breakdown_service,
    entities_service,
)

from zou.app.utils import events


class CastingCsvImportResource(BaseCsvProjectImportResource):
    def post(self, project_id):
        """
        Import project casting links via a .csv file.
        ---
        tags:
          - Import
        consumes:
          - multipart/form-data
        parameters:
          - in: path
            name: project_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: formData
            name: file
            type: file
            required: true
        responses:
            201:
                description: The lists of imported casting links.
            400:
                description: The .csv file is not properly formatted.
        """
        return super().post(project_id)

    def prepare_import(self, project_id):
        self.asset_type_map = {}
        self.asset_map = {}
        self.episode_map = {}
        self.episode_name_map = {}
        self.sequence_map = {}
        self.shot_map = {}

        asset_types = assets_service.get_asset_types()
        for asset_type in asset_types:
            self.asset_type_map[asset_type["id"]] = slugify(asset_type["name"])

        assets = assets_service.get_assets({"project_id": project_id})
        for asset in assets:
            key = self.get_asset_key(asset)
            self.asset_map[key] = asset["id"]

        project = projects_service.get_project(project_id)
        self.is_tv_show = projects_service.is_tv_show(project)
        if self.is_tv_show:
            episodes = shots_service.get_episodes({"project_id": project_id})
            for episode in episodes:
                self.episode_map[episode["id"]] = slugify(episode["name"])
                key = self.get_episode_key(episode)
                self.episode_name_map[key] = episode["id"]

        sequences = shots_service.get_sequences({"project_id": project_id})
        for sequence in sequences:
            key = self.get_sequence_key(sequence)
            self.sequence_map[sequence["id"]] = key

        shots = shots_service.get_shots({"project_id": project_id})
        for shot in shots:
            key = self.get_shot_key(shot)
            self.shot_map[key] = shot["id"]

    def get_asset_key(self, asset):
        asset_type_name = self.asset_type_map[asset["entity_type_id"]]
        return f"{asset_type_name}{slugify(asset['name'])}"

    def get_sequence_key(self, sequence):
        episode_name = ""
        if sequence["parent_id"] in self.episode_map:
            episode_name = self.episode_map[sequence["parent_id"]]
        return f"{episode_name}{slugify(sequence['name'])}"

    def get_shot_key(self, shot):
        sequence_key = self.sequence_map[shot["parent_id"]]
        return f"{sequence_key}{slugify(shot['name'])}"

    def get_episode_key(self, episode):
        return f"episode{slugify(episode['name'])}"

    def import_row(self, row, project_id):
        asset_key = slugify(f"{row['Asset Type']}{row['Asset']}")
        if row.get("Episode") in ["MP", None]:
            row["Episode"] = ""
        target_key = slugify(f"{row['Episode']}{row['Parent']}{row['Name']}")
        occurences = 1
        if len(row["Occurences"]) > 0:
            occurences = int(row["Occurences"])

        asset_id = self.asset_map.get(asset_key, None)
        target_id = self.shot_map.get(target_key, None)
        if target_id is None:
            target_id = self.asset_map.get(
                slugify(f"{row['Parent']}{row['Name']}"), None
            )
            if target_id is None:
                target_id = self.episode_name_map.get(target_key, None)

        label = slugify(row.get("Label", "fixed"))

        if asset_id is not None and target_id is not None:
            entity = entities_service.get_entity_raw(target_id)
            link = breakdown_service.get_entity_link_raw(target_id, asset_id)
            if link is None:
                breakdown_service.create_casting_link(
                    target_id, asset_id, occurences, label
                )
                entity.update({"nb_entities_out": entity.nb_entities_out + 1})
            else:
                link.update({"nb_occurences": occurences, "label": label})
            entity_id = str(entity.id)
            if shots_service.is_shot(entity.serialize()):
                breakdown_service.refresh_shot_casting_stats(
                    entity.serialize()
                )
                events.emit(
                    "shot:casting-update",
                    {
                        "shot_id": entity_id,
                        "nb_entities_out": entity.nb_entities_out,
                    },
                    project_id=str(entity.project_id),
                )
            elif shots_service.is_episode(entity.serialize()):
                events.emit(
                    "episode:casting-update",
                    {
                        "episode_id": entity_id,
                        "nb_entities_out": entity.nb_entities_out,
                    },
                    project_id=str(entity.project_id),
                )
            else:
                events.emit(
                    "asset:casting-update",
                    {"asset_id": entity_id},
                    project_id=str(entity.project_id),
                )



"""add people salary scale table

Revision ID: 45f739ef962a
Revises: 2762a797f1f9
Create Date: 2025-04-02 10:11:12.544243

"""

from alembic import op
import sqlalchemy as sa
import sqlalchemy_utils
import sqlalchemy_utils
import uuid

# revision identifiers, used by Alembic.
revision = "45f739ef962a"
down_revision = "2762a797f1f9"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "salary_scale",
        sa.Column(
            "department_id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=False,
        ),
        sa.Column("senior_supervisor_salary", sa.Integer(), nullable=False),
        sa.Column("mid_supervisor_salary", sa.Integer(), nullable=False),
        sa.Column("junior_supervisor_salary", sa.Integer(), nullable=False),
        sa.Column("senior_lead_salary", sa.Integer(), nullable=False),
        sa.Column("mid_lead_salary", sa.Integer(), nullable=False),
        sa.Column("junior_lead_salary", sa.Integer(), nullable=False),
        sa.Column("senior_artist_salary", sa.Integer(), nullable=False),
        sa.Column("mid_artist_salary", sa.Integer(), nullable=False),
        sa.Column("junior_artist_salary", sa.Integer(), nullable=False),
        sa.Column(
            "id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=False,
        ),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("updated_at", sa.DateTime(), nullable=True),
        sa.ForeignKeyConstraint(
            ["department_id"],
            ["department.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    with op.batch_alter_table("salary_scale", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_salary_scale_department_id"),
            ["department_id"],
            unique=False,
        )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("salary_scale", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_salary_scale_department_id"))

    op.drop_table("salary_scale")
    # ### end Alembic commands ###



from zou.app import db
from zou.app.models.serializer import SerializerMixin
from zou.app.models.base import BaseMixin


class CustomAction(db.Model, BaseMixin, SerializerMixin):
    """
    Custom actions are HTTP links that can be activated outside of the API.
    They are mainly aimed at being used by the web frontend to allow studio
    to make custom HTTP calls.
    """

    name = db.Column(db.String(80), nullable=False)
    url = db.Column(db.String(400))
    entity_type = db.Column(db.String(40), default="all")
    is_ajax = db.Column(db.Boolean(), default=False)



from flask import abort
from flask_jwt_extended import jwt_required

from flask_restful import Resource
from zou.app.utils import csv_utils, permissions


class BaseCsvExport(Resource):
    def __init__(self):
        Resource.__init__(self)
        self.file_name = "export"

    def check_permissions(self):
        permissions.check_manager_permissions()
        return True

    def prepare_import(self):
        pass

    @jwt_required()
    def get(self):
        """
        Export as csv.
        ---
        tags:
            - Export
        responses:
            200:
                description: Exported as csv
        """
        self.prepare_import()
        try:
            self.check_permissions()
            csv_content = []
            csv_content.append(self.build_headers())
            results = self.build_query().all()
            for result in results:
                csv_content.append(self.build_row(result))
        except permissions.PermissionDenied:
            abort(403)

        return csv_utils.build_csv_response(
            csv_content, file_name=self.file_name
        )
